
"""Module for holding functions for calculating physical quantities for the physical rotation and activity model."""

# Imports for standard stuff needed here
import numpy as np
import sys as sys
import copy as copy

# Imports for mors modules
import mors.miscellaneous as misc
import mors.stellarevo as SE
import mors.constants as const
import mors.parameters as params

def dOmegadt(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns rates of change of rotation.
    
    This is the main function to be used by the solvers to get the rates of change of the core and envelope 
    rotation rates given basic stellar parameters. The actual calculations of these are done in RotationQuantities() 
    which itself also returns a very large number of other things used during the calculation of the rates 
    of change.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    OmegaEnv : float
        Envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaCore : float
        Core rotation rate in OmegaSun (=2.67e-6 rad/s).
    params : dict , optional
        Dictionary holding model parameters. 
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    dOmegaEnvdt : float
        Rate of change of envelope rotation rate in OmegaSun Myr^-1.
    dOmegaCoredt : float
        Rate of change of core rotation rate in OmegaSun Myr^-1.
    
    """
    
    # Make sure parameters are set
    if Mstar is None: 
        misc._PrintErrorKill("argument Mstar must be set in call to function")
    if Age is None: 
        misc._PrintErrorKill("argument Age must be set in call to function")
    if OmegaEnv is None: 
        misc._PrintErrorKill("argument OmegaEnv must be set in call to function")
    if OmegaCore is None: 
        misc._PrintErrorKill("argument OmegaCore must be set in call to function")
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Get state of system 
    StarState = RotationQuantities(Mstar=Mstar,Age=Age,OmegaEnv=OmegaEnv,OmegaCore=OmegaCore,params=params,StarEvo=StarEvo)
    
    return ( StarState['dOmegaEnvdt'] , StarState['dOmegaCoredt'] )

def RotationQuantities(Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns rotation quantities including rates of change of rotation.
    
    This is the main function to be used to get the rates of change of the core and envelope rotation 
    rates given basic stellar parameters. The actual calculations of these are done in GetState() which
    itself also returns a very large number of other things used during the calculation of the rates 
    of change.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    OmegaEnv : float
        Envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaCore : float
        Core rotation rate in OmegaSun (=2.67e-6 rad/s).
    params : dict , optional
        Dictionary holding model parameters. 
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    StarState : dict
        Set of rotation quantities.
    
    """
    
    # Make sure parameters are set
    if Mstar is None: 
        misc._PrintErrorKill("argument Mstar must be set in call to function")
    if Age is None: 
        misc._PrintErrorKill("argument Age must be set in call to function")
    if OmegaEnv is None: 
        misc._PrintErrorKill("argument OmegaEnv must be set in call to function")
    if OmegaCore is None: 
        misc._PrintErrorKill("argument OmegaCore must be set in call to function")
        
    #--------------------------------------------------------
    
    # Setup stellar parameter dictionary
    # This will be used to hold all of the 
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Start dictionary
    StarState = {}
    StarStateUnits = {}
    
    # Add basic input parameters 
    StarState['Mstar'] = Mstar
    StarState['Age'] = Age
    StarState['OmegaEnv'] = OmegaEnv
    StarState['OmegaCore'] = OmegaCore
    
    # Get rotation period from surface rotation in days
    StarState['Prot'] = _Prot(OmegaEnv)
    
    # Radius in cm
    StarState['Rstar'] = StarEvo.Rstar(Mstar,Age)
    
    # Convective turnover time in days
    StarState['tauConv'] = StarEvo.tauConv(Mstar,Age)
    
    # Rossby number
    StarState['Ro'] = _Ro(StarState)
    
    # Get moments of interia in g cm^2
    StarState['Itotal'] = StarEvo.Itotal(Mstar,Age)
    StarState['Ienv'] = StarEvo.Ienv(Mstar,Age)
    StarState['Icore'] = StarEvo.Icore(Mstar,Age)
    
    # Rates of change of moment of inertia in g cm^2 s^-1
    StarState['dItotaldt'] = StarEvo.dItotaldt(Mstar,Age) / const.Myr
    StarState['dIenvdt'] = StarEvo.dIenvdt(Mstar,Age) / const.Myr
    StarState['dIcoredt'] = StarEvo.dIcoredt(Mstar,Age) / const.Myr
    
    # Rcore in cm
    StarState['Rcore'] = StarEvo.Rcore(Mstar,Age) * const.Rsun
    
    # dMcoredt in g s^-1
    StarState['dMcoredt'] = StarEvo.dMcoredt(Mstar,Age) * const.Msun / const.Myr
    
    # Mass loss rate in g s^-1
    StarState['Mdot'] = _Mdot(StarState,params=params) * const.Msunyr_
    
    # Get dipole field strength in G
    StarState['Bdip'] = _Bdip(StarState,params=params)
    
    # Get surface escape velocity in cm s^-1
    StarState['vEsc'] = _vEsc(Mstar,StarState['Rstar'])
    
    #--------------------------------------------------------
    
    # Determine if doing core-envelope decoupling
    CoreEnvelopeDecoupling = _shouldCoreEnvelopeDecoupling( StarState , params=params )
    
    #--------------------------------------------------------
    
    # Add moment of interia change torque
    if params['MomentInertiaChangeTorque']:
        StarState['torqueEnvMom'] , StarState['torqueCoreMom'] = _torqueMoment( StarState , CoreEnvelopeDecoupling )
    else:
        StarState['torqueEnvMom'] = StarState['torqueCoreMom'] = 0.0
    
    # Add core-growth torque
    if params['CoreGrowthTorque'] and CoreEnvelopeDecoupling:
        StarState['torqueEnvCG'] , StarState['torqueCoreCG'] = _torqueCoreGrowth( StarState )
    else:
        StarState['torqueEnvCG'] = StarState['torqueCoreCG'] = 0.0
    
    # Add wind spin-down torque
    if params['WindTorque']:
        StarState['torqueEnvWind'] = _torqueWind( StarState , params=params )
    else:
        StarState['torqueEnvWind'] = 0.0
    
    # Add wind spin-down torque
    if params['CoreEnvelopeTorque'] and CoreEnvelopeDecoupling:
        StarState['torqueEnvCE'] , StarState['torqueCoreCE'] = _torqueCoreEnvelope( StarState , params=params )
    else:
        StarState['torqueEnvCE'] = StarState['torqueCoreCE'] = 0.0
    
    # Add disk locking torque
    if params['DiskLocking']:
        StarState['torqueEnvDL'] = _torqueDiskLocking( StarState , CoreEnvelopeDecoupling , params=params )
    else:
        StarState['torqueEnvDL']  = 0.0
    
    #--------------------------------------------------------
    # Envelope rotation rate of change
    
    # Get total torque on envelope
    StarState['torqueEnv'] = StarState['torqueEnvMom'] + StarState['torqueEnvCG'] + StarState['torqueEnvWind'] + StarState['torqueEnvCE'] + StarState['torqueEnvDL']
    
    # Get rates of change of rotation in rad s^-2
    StarState['dOmegaEnvdt'] = StarState['torqueEnv'] / StarState['Ienv']
    
    # Get envelope rates of change in OmegaSun Myr^-1
    StarState['dOmegaEnvdt'] *= const.Myr / const.OmegaSun
    
    #--------------------------------------------------------
    # Core rotation rate of change
    
    # Have core spin-up with envelope if not developed yet
    if CoreEnvelopeDecoupling:
        
        # Get total torque on core
        StarState['torqueCore'] = StarState['torqueCoreMom'] + StarState['torqueCoreCG'] + StarState['torqueCoreCE']
        
        # Get rates of change of rotation in rad s^-2
        StarState['dOmegaCoredt'] = StarState['torqueCore'] / StarState['Icore']
        
        # Get envelope rates of change in OmegaSun Myr^-1
        StarState['dOmegaCoredt'] *= const.Myr / const.OmegaSun
        
    else:
        
        # There should be no torque on the core and it should spin up or down with the envelope
        StarState['torqueCore'] = 0.0
        StarState['dOmegaCoredt'] = StarState['dOmegaEnvdt']
    
    #--------------------------------------------------------
    
    return StarState

def ExtendedQuantities(StarState=None,Mstar=None,Age=None,OmegaEnv=None,OmegaCore=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns extended set of physical quantities not calculated by RotationQuantities.
    
    This is the main function to be used to get things like stellar XUV emission. The user can either input as arguments the 
    dictionary StarState, which is returned by RotationQuantities, or the basic stellar parameters Mstar, Age, OmegaEnv, and
    OmegaCore (final one not necessary). If StarState is not input, this function will first call RotationQuantities to get it.
    Either way, the return will be the same dictionary with additional parameters added to it.
    
    Parameters
    ----------
    StarState : dict , optional
        Set of quantities already calculated by RotationQuantities().
    Mstar : float , optional
        Mass of star in Msun.
    Age : float , optional
        Age of star.
    OmegaEnv : float 
        Envelope rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaCore : float 
        Core rotation rate in OmegaSun (=2.67e-6 rad/s).
    params : dict , optional
        Dictionary holding model parameters. 
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    StarState : dict
        Set of extended quantities.
    
    """
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
        
    # Get StarState if it was not already set
    if StarState is None: 
        
        # Make sure required parameters were set
        if Mstar is None: 
            misc._PrintErrorKill("argument Mstar must be set in call to function")
        if Age is None: 
            misc._PrintErrorKill("argument Age must be set in call to function")
        if OmegaEnv is None: 
            misc._PrintErrorKill("argument OmegaEnv must be set in call to function")
        if OmegaCore is None: 
            misc._PrintErrorKill("argument OmegaCore must be set in call to function")
        
        # Get rotation quantities
        StarState = RotationQuantities(Mstar=Mstar,Age=Age,OmegaEnv=OmegaEnv,OmegaCore=OmegaCore,params=params,StarEvo=StarEvo)
        
    # Get Lbol in erg s^-1
    StarState['Lbol'] = StarEvo.Lbol(Mstar,Age) * const.LbolSun
    
    # Get effective temperature in K
    StarState['Teff'] = StarEvo.Teff(Mstar,Age)
    
    # Get Lx in erg s^-1, Fx in erg s^-1 cm^-2, and Rx
    StarState['Lx'] , StarState['Fx'] , StarState['Rx'] = _Xray(StarState,params=params)
    
    # Get Tcor in MK
    StarState['Tcor'] = _Tcor(StarState,params=params)
    
    # Get Leuv1 in erg s^-1, Feuv1 in erg s^-1 cm^-2, and Reuv1 (10-36 nm)
    StarState['Leuv1'] , StarState['Feuv1'] , StarState['Reuv1'] = _EUV1(StarState,params=params)
    
    # Get Leuv2 in erg s^-1, Feuv2 in erg s^-1 cm^-2, and Reuv2 (36-92 nm)
    StarState['Leuv2'] , StarState['Feuv2'] , StarState['Reuv2'] = _EUV2(StarState,params=params)
    
    # Get Leuv in erg s^-1, Feuv in erg s^-1 cm^-2, and Reuv
    StarState['Leuv'] , StarState['Feuv'] , StarState['Reuv'] = _EUV(StarState,params=params)
    
    # Get Lly in erg s^-1, Fly in erg s^-1 cm^-2, and Rly
    StarState['Lly'] , StarState['Fly'] , StarState['Rly'] = _Lymanalpha(StarState,params=params)
    
    # Get habitable zone fluxes in erg s^-1 cm^-2
    aOrbHZ2 = aOrbHZ( Mstar=StarState['Mstar'] , params=params )
    StarState['FxHZ'] = StarState['Lx'] / ( 4.0 * const.Pi * (aOrbHZ2['HZ']*const.AU)**2.0 )
    StarState['Feuv1HZ'] = StarState['Leuv1'] / ( 4.0 * const.Pi * (aOrbHZ2['HZ']*const.AU)**2.0 )
    StarState['Feuv2HZ'] = StarState['Leuv2'] / ( 4.0 * const.Pi * (aOrbHZ2['HZ']*const.AU)**2.0 )
    StarState['FeuvHZ'] = StarState['Leuv'] / ( 4.0 * const.Pi * (aOrbHZ2['HZ']*const.AU)**2.0 )
    StarState['FlyHZ'] = StarState['Lly'] / ( 4.0 * const.Pi * (aOrbHZ2['HZ']*const.AU)**2.0 )
    
    return StarState 

def QuantitiesUnits(StarState=None):
    """Returns dictionary of units for each physical quantity in calculated by RotationQuantities() and ExtendedQuantities().
    
    For output purposes, this is essential since it gives the user a way to know the units of the quantities output. It is 
    recommended to input also 'StarState' which will cause the function to also check that all quantities in StarState do
    in fact have units and will cause the function to only return units for quantities that are included, so for instance if
    StarState was only calculated by RotationQuantities() and not by ExtendedQuantities(), then only units for the rotation
    quantities will be included.
    
    Parameters
    ----------
    StarState : dict , optional
        Set of quantities already calculated by RotationQuantities() and possibly ExtendedQuantities().
    
    Returns
    ----------
    StarStateUnits : dict
        Dictionary of strings holding units for quantities.
    
    """
    
    # Create empty dictionary
    StarStateUnits = {}
    
    # Get initial dictionary of units
    StarStateUnits['Mstar'] = 'Msun'
    StarStateUnits['dAge'] = 'Myr'
    StarStateUnits['Age'] = 'Myr'
    StarStateUnits['OmegaEnv'] = 'OmegaSun =2.67e-6 rad s^-1'
    StarStateUnits['OmegaCore'] = 'OmegaSun =2.67e-6 rad s^-1'  
    StarStateUnits['Prot'] = 'days'
    StarStateUnits['Rstar'] = 'Rsun'
    StarStateUnits['tauConv'] = 'days'
    StarStateUnits['Ro'] = ''
    StarStateUnits['Itotal'] = 'g cm^2'
    StarStateUnits['Ienv'] = 'g cm^2'
    StarStateUnits['Icore'] = 'g cm^2'  
    StarStateUnits['dItotaldt'] = 'g cm^2 s^-1'
    StarStateUnits['dIenvdt'] = 'g cm^2 s^-1'
    StarStateUnits['dIcoredt'] = 'g cm^2 s^-1'
    StarStateUnits['Rcore'] = 'Rsun'
    StarStateUnits['dMcoredt'] = 'g s^-1'
    StarStateUnits['Mdot'] = 'g s^-1'
    StarStateUnits['Bdip'] = 'G'
    StarStateUnits['vEsc'] = 'cm s^-1'
    StarStateUnits['torqueEnvMom'] = 'erg'
    StarStateUnits['torqueCoreMom'] = 'erg'
    StarStateUnits['torqueEnvCG'] = 'erg'
    StarStateUnits['torqueCoreCG'] = 'erg'
    StarStateUnits['torqueEnvWind'] = 'erg'
    StarStateUnits['torqueEnvCE'] = 'erg'
    StarStateUnits['torqueCoreCE'] = 'erg'
    StarStateUnits['torqueEnvDL'] = 'erg'
    StarStateUnits['torqueEnv'] = 'erg'
    StarStateUnits['dOmegaEnvdt'] = 'OmegaSun Myr^-1'
    StarStateUnits['torqueCore'] = 'erg'
    StarStateUnits['dOmegaCoredt'] = 'OmegaSun Myr^-1'
    StarStateUnits['Lbol'] = 'erg s^-1'
    StarStateUnits['Teff'] = 'K'
    StarStateUnits['Lx'] = 'erg s^-1'
    StarStateUnits['Fx'] = 'erg s^-1 cm^-2'
    StarStateUnits['Rx'] = ''
    StarStateUnits['Tcor'] = 'MK'
    StarStateUnits['Leuv1'] = 'erg s^-1'
    StarStateUnits['Feuv1'] = 'erg s^-1 cm^-2'
    StarStateUnits['Reuv1'] = ''
    StarStateUnits['Leuv2'] = 'erg s^-1'
    StarStateUnits['Feuv2'] = 'erg s^-1 cm^-2'
    StarStateUnits['Reuv2'] = ''
    StarStateUnits['Leuv'] = 'erg s^-1'
    StarStateUnits['Feuv'] = 'erg s^-1 cm^-2'
    StarStateUnits['Reuv'] = ''  
    StarStateUnits['Lly'] = 'erg s^-1'
    StarStateUnits['Fly'] = 'erg s^-1 cm^-2'
    StarStateUnits['Rly'] = ''
    StarStateUnits['FxHZ'] = 'erg s^-1 cm^-2'
    StarStateUnits['Feuv1HZ'] = 'erg s^-1 cm^-2'
    StarStateUnits['Feuv2HZ'] = 'erg s^-1 cm^-2'
    StarStateUnits['FeuvHZ'] = 'erg s^-1 cm^-2'
    StarStateUnits['FlyHZ'] = 'erg s^-1 cm^-2'
        
    # If StarState was set in call to function then only include units for quantities in StarState and
    # make sure all quantities in StarState have units given (even if the quantity is dimensionless)
    if not StarState is None:
        
        # Make deep copy of StarStateUnits
        StarStateUnitsTemp = copy.deepcopy(StarStateUnits)
        
        # Start new StarStateUnits
        StarStateUnits = {}
        
        # Loop over quantities in StarState and add each one to StarStateUnits
        for quantity in StarState:
            
            # Make sure it is included in list of units above
            if not ( quantity in StarStateUnitsTemp ):
                misc._PrintErrorKill("units for "+quantity+" not included in units list")
            
            # Add to dictionary
            StarStateUnits[quantity] = StarStateUnitsTemp[quantity]
    
    return StarStateUnits

def Lxuv(Mstar=None,Age=None,Omega=None,OmegaEnv=None,Prot=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns X-ray, EUV, and Ly-alpha luminosities in erg s^-1.
    
    The user can supply a mass, age, and surface rotation rate for a star and it returns the XUV luminosities as 
    a dictionary. The rotation rate can be either the rotation velocity as a multiple of the solar rotation 
    rate (2.67e-6 rad s^-1) using either Omega or OmegaEnv keyword arguments, or as a rotation period in days using 
    the Prot keyword argument, The user can optionally also give a parameter dictionary and an instance of the 
    StarEvo class, though this is not necessary and if these are not specified then the defaults will be used.
    
    Parameters
    ----------
    Band : str
        Wavelength band to calculate.
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    Omega : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaEnv : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    Prot : float , optional
        Surface rotation period in days.
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    LxuvDict : dict
        Dictionary of luminositie in erg s^-1.
    
    """
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Make sure mass and age are specified
    if Mstar is None:
        misc._PrintErrorKill("Mstar must be set in call to function")
    if Age is None:
        misc._PrintErrorKill("Age must be set in call to function")
    
    # Make sure rotation is set
    if ( Omega is None ) and ( OmegaEnv is None ) and ( Prot is None ):
        misc._PrintErrorKill("Omega, OmegaEnv, or Prot must be set in call to function")
        
    # Make sure only one rotation rate is set (add up number of set parameters)
    nSet = 0
    if not Omega is None:
        nSet += 1
    if not OmegaEnv is None:
        nSet += 1
    if not Prot is None:
        nSet += 1
    if not ( nSet == 1 ):
        misc._PrintErrorKill("can only set one of Omega, OmegaEnv, and Prot in call to function")
    
    # Get Omega not set
    if not OmegaEnv is None:
        Omega = OmegaEnv
    if not Prot is None:
        Omega = _Omega(Prot)
        
    # The function _Xray needs Ro, Rstar, and Lbol so make a StarState dictionary with these
    StarState = {}
    StarState['OmegaEnv'] = Omega
    StarState['Rstar'] = StarEvo.Rstar(Mstar,Age)
    StarState['Prot'] = _Prot(Omega)
    StarState['tauConv'] = StarEvo.tauConv(Mstar,Age)
    StarState['Ro'] = _Ro(StarState)
    StarState['Lbol'] = StarEvo.Lbol(Mstar,Age) * const.LbolSun
    
    # Get X-ray
    StarState['Lx'] , StarState['Fx'] , StarState['Rx'] = _Xray(StarState,params=params)
    
    # Get EUV
    StarState['Leuv1'] , StarState['Feuv1'] , StarState['Reuv1'] = _EUV1(StarState,params=params)
    StarState['Leuv2'] , StarState['Feuv2'] , StarState['Reuv2'] = _EUV2(StarState,params=params)
    StarState['Leuv'] , StarState['Feuv'] , StarState['Reuv'] = _EUV(StarState,params=params)
    
    # Get Ly-alpha 
    StarState['Lly'] , StarState['Fly'] , StarState['Rly'] = _Lymanalpha(StarState,params=params)
    
    # Make dictionary with luminosities
    LxuvDict = {}
    LxuvDict['Lxuv'] = StarState['Lx'] + StarState['Leuv']
    LxuvDict['Lx'] = StarState['Lx']
    LxuvDict['Leuv1'] = StarState['Leuv1']
    LxuvDict['Leuv2'] = StarState['Leuv2']
    LxuvDict['Leuv'] = StarState['Leuv']
    LxuvDict['Lly'] = StarState['Lly']
    
    LxuvDict['Fxuv'] = StarState['Fx'] + StarState['Feuv']
    LxuvDict['Fx'] = StarState['Fx']
    LxuvDict['Feuv1'] = StarState['Feuv1']
    LxuvDict['Feuv2'] = StarState['Feuv2']
    LxuvDict['Feuv'] = StarState['Feuv']
    LxuvDict['Fly'] = StarState['Fly']
    
    LxuvDict['Rxuv'] = StarState['Rx'] + StarState['Reuv']
    LxuvDict['Rx'] = StarState['Rx']
    LxuvDict['Reuv1'] = StarState['Reuv1']
    LxuvDict['Reuv2'] = StarState['Reuv2']
    LxuvDict['Reuv'] = StarState['Reuv']
    LxuvDict['Rly'] = StarState['Rly']
    
    return LxuvDict

def Lx(Mstar=None,Age=None,Omega=None,OmegaEnv=None,Prot=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns X-ray luminosity in erg s^-1.
    
    The user can supply a mass, age, and surface rotation rate for a star and it returns the X-ray luminosity. The
    rotation rate can be either the rotation velocity as a multiple of the solar rotation rate (2.67e-6 rad s^-1) 
    using either Omega or OmegaEnv keyword arguments, or as a rotation period in days using the Prot keyword 
    argument, The user can optionally also give a parameter dictionary and an instance of the StarEvo class, though 
    this is not necessary and if these are not specified then the defaults will be used.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    Omega : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaEnv : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    Prot : float , optional
        Surface rotation period in days.
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    Lx : float
        X-ray luminosity in erg s^-1.
    
    """
    
    # Get dictionary with all luminosities
    LxuvDict = Lxuv(Mstar=Mstar,Age=Age,Omega=Omega,OmegaEnv=OmegaEnv,Prot=Prot,params=params,StarEvo=StarEvo)
    
    return LxuvDict['Lx']

def _Xray(StarState,params=params.paramsDefault):
    """Takes star state, returns Lx in erg s^-1, Fx in erg s^-1 cm^-1, and Rx."""
    
    # Get Rx depending on Ro and regime
    if ( StarState['Ro'] < params['RoSatXray'] ):
        Rx = params['C1Xray'] * StarState['Ro']**params['beta1Xray']
    else:
        Rx = params['C2Xray'] * StarState['Ro']**params['beta2Xray']
    
    # Get Lx from that
    Lx = StarState['Lbol'] * Rx
    
    # Get Fx
    Fx = Lx / ( 4.0 * const.Pi * (StarState['Rstar']*const.Rsun)**2.0 )
    
    return Lx , Fx , Rx

def XrayScatter(XrayAverage,params=params.paramsDefault):
    """Takes stellar X-ray, returns X-ray scatter around this value.
    
    The other functions in this code calculate an average Lx for a star given its parameters, making Lx a unique
    function of mass, age, and rotation. In reality there is a scatter around these values that appear somewhat
    random, likely due to variability. This scatter can be described as a log normal probability density function.
    This function can be used to get values for the scatter for a given stellar Lx. Specifically, if LxAverage is
    the average mass, age, and omega dependent X-ray luminosity, the true Lx is given by LxAverage + deltaLx, and
    it is this deltaLx that this function calculates. This function works equally well for inputting Rx or Fx.
    
    Parameters
    ----------
    XrayAverage : float or numpy.ndarray
        Values for stellar Lx, Fx, or Rx in any units.
    params : dict , optional
        Dictionary holding model parameters.
    
    Returns
    ----------
    deltaXray : float or numpy.ndarray
        Values for stellar deltaLx, deltaFx, or deltaRx in input units.
    
    """
    
    # Get number of stars
    if isinstance(XrayAverage,(float,int)):
        nStars = 1
    else:
        nStars = len(XrayAverage)
    
    # Get random number from normal distribution
    rand = np.random.normal( 0.0 , params['sigmaXray'] , nStars )
    
    # Add random to the log of the quantities to get scattered values
    XrayAverageScattered = 10.0**( np.log10(XrayAverage) + rand )
    
    # Get change
    deltaXray = XrayAverageScattered - XrayAverage
    
    return deltaXray

def XUVScatter(XUVAverage,params=params.paramsDefault):
    """Takes stellar XUV values, returns scatter around these value.
    
    The other functions in this code calculate an average XUV values for a star given its parameters, making them unique
    functions of mass, age, and rotation. In reality there is a scatter around these values that appear somewhat
    random, likely due to variability. This scatter can be described as a log normal probability density function.
    This function can be used to get values for the scatter for a given stellar Lx. Specifically, if LxAverage is
    the average mass, age, and omega dependent X-ray luminosity, the true Lx is given by LxAverage + deltaLx, and
    it is this deltaLx that this function calculates. This function calculates these scatter values for all XUV
    parameters calculated by Lxuv above
    
    Parameters
    ----------
    XUVAverage : dict
        Dictionary of values returned by Lxuv().
    params : dict , optional
        Dictionary holding model parameters.
    
    Returns
    ----------
    deltaXUV : dict
        Values for deltaLx, deltaFx, etc. for all quantities calculated by Lxuv().
    
    """
    
    
    # Get random number for X-rays from normal distribution
    randXray = np.random.normal( 0.0 , params['sigmaXray'] )
    
    # Get random numbers for other quantities (not including composite like EUV=EUV1+EUV2)
    randEUV1 = 0.681 * randXray
    randEUV2 = 0.920 * randEUV1
    randLy = 0.375 * randXray
    
    # Add random to the log of the quantities to get scattered values
    LxScattered = 10.0**( np.log10(XUVAverage['Lx']) + randXray )
    Leuv1Scattered = 10.0**( np.log10(XUVAverage['Leuv1']) + randEUV1 )
    Leuv2Scattered = 10.0**( np.log10(XUVAverage['Leuv2']) + randEUV2 )
    LlyScattered = 10.0**( np.log10(XUVAverage['Lly']) + randLy )
    
    FxScattered = 10.0**( np.log10(XUVAverage['Fx']) + randXray )
    Feuv1Scattered = 10.0**( np.log10(XUVAverage['Feuv1']) + randEUV1 )
    Feuv2Scattered = 10.0**( np.log10(XUVAverage['Feuv2']) + randEUV2 )
    FlyScattered = 10.0**( np.log10(XUVAverage['Fly']) + randLy )
    
    RxScattered = 10.0**( np.log10(XUVAverage['Rx']) + randXray )
    Reuv1Scattered = 10.0**( np.log10(XUVAverage['Reuv1']) + randEUV1 )
    Reuv2Scattered = 10.0**( np.log10(XUVAverage['Reuv2']) + randEUV2 )
    RlyScattered = 10.0**( np.log10(XUVAverage['Rly']) + randLy )
    
    # Get scattered values for composite quantities
    LeuvScattered = Leuv1Scattered + Leuv2Scattered
    FeuvScattered = Feuv1Scattered + Feuv2Scattered
    ReuvScattered = Reuv1Scattered + Reuv2Scattered
    
    LxuvScattered = LxScattered + LeuvScattered
    FxuvScattered = FxScattered + FeuvScattered
    RxuvScattered = RxScattered + ReuvScattered
    
    # Get changes in array
    deltaXUV = {}
    
    deltaXUV['Lxuv'] = LxuvScattered - XUVAverage['Lxuv']
    deltaXUV['Lx'] = LxScattered - XUVAverage['Lx']
    deltaXUV['Leuv'] = LeuvScattered - XUVAverage['Leuv']
    deltaXUV['Leuv1'] = Leuv1Scattered - XUVAverage['Leuv1']
    deltaXUV['Leuv2'] = Leuv2Scattered - XUVAverage['Leuv2']
    deltaXUV['Lly'] = LlyScattered - XUVAverage['Lly']
    
    deltaXUV['Fxuv'] = FxuvScattered - XUVAverage['Fxuv']
    deltaXUV['Fx'] = FxScattered - XUVAverage['Fx']
    deltaXUV['Feuv'] = FeuvScattered - XUVAverage['Feuv']
    deltaXUV['Feuv1'] = Feuv1Scattered - XUVAverage['Feuv1']
    deltaXUV['Feuv2'] = Feuv2Scattered - XUVAverage['Feuv2']
    deltaXUV['Fly'] = FlyScattered - XUVAverage['Fly']
    
    deltaXUV['Rxuv'] = RxuvScattered - XUVAverage['Rxuv']
    deltaXUV['Rx'] = RxScattered - XUVAverage['Rx']
    deltaXUV['Reuv'] = ReuvScattered - XUVAverage['Reuv']
    deltaXUV['Reuv1'] = Reuv1Scattered - XUVAverage['Reuv1']
    deltaXUV['Reuv2'] = Reuv2Scattered - XUVAverage['Reuv2']
    deltaXUV['Rly'] = RlyScattered - XUVAverage['Rly']
    
    return deltaXUV

def _Tcor(StarState,params=params.paramsDefault):
    """Takes star state, returns average coronal temperature in MK."""
    
    # Get Tcor depending on Fx using equations from Johnstone & Guedel (2015)
    Tcor = 0.11 * StarState['Fx']**0.26
    
    return Tcor

def Leuv(Mstar=None,Age=None,Omega=None,OmegaEnv=None,Prot=None,band=0,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns EUV luminosity in erg s^-1.
    
    The user can supply a mass, age, and surface rotation rate for a star and it returns the EUV luminosity. 
    By default, this will be the luminosity in the 10-92 nm band, but using the band keyword argument, the 
    user can set specify either the 10-36 nm range with 'band=1' or the 36-92 nm range for 'band=2'. The default
    10-92 nm range is set with 'band=0'. The rotation rate can be either the rotation velocity as a multiple 
    of the solar rotation rate (2.67e-6 rad s^-1) using either Omega or OmegaEnv keyword arguments, or as a 
    rotation period in days using the Prot keyword argument, The user can optionally also give a parameter 
    dictionary and an instance of the StarEvo class, though this is not necessary and if these are not specified 
    then the defaults will be used.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    Omega : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaEnv : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    Prot : float , optional
        Surface rotation period in days.
    band : float , optional
        Which wavelength band to return (=0 for 10-92 nm; =1 for 10-32 nm; =2 for 32-92 nm).
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    Leuv : float
        EUV luminosity in erg s^-1.
    
    """
    
    # Get dictionary with all luminosities
    LxuvDict = Lxuv(Mstar=Mstar,Age=Age,Omega=Omega,OmegaEnv=OmegaEnv,Prot=Prot,params=params,StarEvo=StarEvo)
    
    # Return depending on the desired band
    if ( band == 0 ):
        return LxuvDict['Leuv']
    elif ( band == 1 ):
        return LxuvDict['Leuv1']
    elif ( band == 2 ):
        return LxuvDict['Leuv2']
    else:
        misc._PrintErrorKill("invalid band set in call to function")
        
    return -1

def _EUV1(StarState,params=params.paramsDefault):
    """Takes star state, returns Leuv1 in erg s^-1, Feuv in erg s^-1 cm^-1, and Reuv (10-36 nm)."""
    
    # Get Feuv from Fx using equation from Johnstone et al. (2020)
    Feuv1 = 10.0**( 2.04 + 0.681 * np.log10(StarState['Fx']) )
    
    # Get Leuv1 from this
    Leuv1 = Feuv1 * ( 4.0 * const.Pi * (StarState['Rstar']*const.Rsun)**2.0 )
    
    # Get Reuv1
    Reuv1 = Leuv1 / StarState['Lbol']
    
    return Leuv1 , Feuv1 , Reuv1

def _EUV2(StarState,params=params.paramsDefault):
    """Takes star state, returns Leuv2 in erg s^-1, Feuv2 in erg s^-1 cm^-1, and Reuv2 (36-92 nm)."""
    
    # Get Feuv2 from Feuv1 using equation from Johnstone et al. (2020)
    Feuv2 = 10.0**( -0.0341 + 0.92 * np.log10(StarState['Feuv1']) )
    
    # Get Leuv1 from this
    Leuv2 = Feuv2 * ( 4.0 * const.Pi * (StarState['Rstar']*const.Rsun)**2.0 )
    
    # Get Reuv2
    Reuv2 = Leuv2 / StarState['Lbol']
    
    return Leuv2 , Feuv2 , Reuv2

def _EUV(StarState,params=params.paramsDefault):
    """Takes star state, returns Leuv in erg s^-1, Feuv in erg s^-1 cm^-1, and Reuv (10-92 nm)."""
    
    # Just add quantities 
    Leuv = StarState['Leuv1'] + StarState['Leuv2']
    Feuv = StarState['Feuv1'] + StarState['Feuv2']
    Reuv = StarState['Reuv1'] + StarState['Reuv2']
    
    return Leuv , Feuv , Reuv

def Lly(Mstar=None,Age=None,Omega=None,OmegaEnv=None,Prot=None,params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns Ly-alpha luminosity in erg s^-1.
    
    The user can supply a mass, age, and surface rotation rate for a star and it returns the Ly-alpha luminosity. The
    rotation rate can be either the rotation velocity as a multiple of the solar rotation rate (2.67e-6 rad s^-1) 
    using either Omega or OmegaEnv keyword arguments, or as a rotation period in days using the Prot keyword 
    argument, The user can optionally also give a parameter dictionary and an instance of the StarEvo class, though 
    this is not necessary and if these are not specified then the defaults will be used.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star.
    Omega : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    OmegaEnv : float , optional
        Surface rotation rate in OmegaSun (=2.67e-6 rad/s).
    Prot : float , optional
        Surface rotation period in days.
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    Lly : float
        Lyman-alpha luminosity in erg s^-1.
    
    """
    
    # Get dictionary with all luminosities
    LxuvDict = Lxuv(Mstar=Mstar,Age=Age,Omega=Omega,OmegaEnv=OmegaEnv,Prot=Prot,params=params,StarEvo=StarEvo)
    
    return LxuvDict['Lly']

def _Lymanalpha(StarState,params=params.paramsDefault):
    """Takes star state, returns Lly in erg s^-1, Fly in erg s^-1 cm^-1, and Rly (10-92 nm)."""
    
    # Get Fly using from Fx using equation from Johnstone et al. (2020)
    Fly = 10.0**( 3.97 + 0.375 * np.log10(StarState['Fx']) )
    
    # Get Lly
    Lly = Fly * ( 4.0 * const.Pi * (StarState['Rstar']*const.Rsun)**2.0 )
    
    # Get Rly
    Rly = Lly / StarState['Lbol']
    
    return Lly , Fly , Rly

def _torqueDiskLocking(StarState,CoreEnvelopeDecoupling,params=params.paramsDefault):
    """Takes stellar parameters, returns disk-locking torque."""
    
    # As it is currently setup, this will only work if the other torques are calculated first
    
    # Get disk locking age
    ageDL = params['aDiskLock'] * StarState['OmegaEnv']**params['bDiskLock']
    ageDL = min( ageDL , params['ageDLmax'] )
    
    # Return nothing if after disk locking age
    if ( StarState['Age'] > ageDL ):
        return 0.0
    
    # Get torque on envelope
    if ( CoreEnvelopeDecoupling ):
        torqueEnvDL = - ( StarState['torqueEnvWind'] + StarState['torqueEnvCE'] + StarState['torqueEnvCG'] + StarState['torqueEnvMom'] )
    else:
        torqueEnvDL = - ( StarState['torqueEnvWind'] + StarState['torqueEnvMom'] )
    
    return torqueEnvDL

def _torqueCoreEnvelope(StarState,params=params.paramsDefault):
    """Takes stellar parameters, returns core-envelope torque."""
        
    # Get deltaJ
    deltaJ = StarState['Icore'] * StarState['Ienv'] / ( StarState['Icore'] + StarState['Ienv'] ) * ( StarState['OmegaCore'] - StarState['OmegaEnv'] ) * const.OmegaSun
    
    # Get core-envelope coupling timescale in Myr
    timeCE = params['aCoreEnvelope'] * StarState['OmegaEnv']**params['bCoreEnvelope'] * StarState['Mstar']**params['cCoreEnvelope']
    timeCE = max( timeCE , params['timeCEmin'] )
    
    # Convert timescale to seconds
    timeCE *= const.Myr
    
    # Get torques
    torqueEnvMoment = deltaJ / timeCE
    torqueCoreMoment = - torqueEnvMoment
    
    return ( torqueEnvMoment , torqueCoreMoment )

def _torqueMoment(StarState,CoreEnvelopeDecoupling):
    """Takes stellar parameters, returns moment of inertia change torque in erg."""
    
    # Note that the torque calculated in this function is not actually a real physical torque
    # but it is convenient to write it in the equations and in the code as if it was. 
    
    # Get torque in erg depending on if core-envelope decoupling is being done
    if ( CoreEnvelopeDecoupling ):
        torqueEnvMoment = - StarState['dIenvdt'] * StarState['OmegaEnv']*const.OmegaSun
        torqueCoreMoment = - StarState['dIcoredt'] * StarState['OmegaCore']*const.OmegaSun
    else:
        torqueEnvMoment = - StarState['dItotaldt'] * StarState['OmegaEnv']*const.OmegaSun
        torqueCoreMoment = 0.0
        
    return ( torqueEnvMoment , torqueCoreMoment )

def _torqueCoreGrowth(StarState):
    """Takes stellar parameters, returns core-growth torque in erg."""
    
    # Get torque in erg
    torqueEnvCG = - 2.0/3.0 * StarState['Rcore']**2.0 * StarState['OmegaEnv']*const.OmegaSun * StarState['dMcoredt']
    torqueCoreCG = - torqueEnvCG
    
    return ( torqueEnvCG , torqueCoreCG )

def _torqueWind(StarState,params=params.paramsDefault):
    """Take stellar parameters, returns wind torque in erg."""

    # Some constants from Matt et al. (2012)
    K1Matt = 1.3
    K2Matt = 0.0506
    mMatt = 0.2177
    
    # Calculate torque
    torqueWind = K1Matt**2.0 * StarState['Bdip']**(4.0*mMatt) * StarState['Mdot']**(1.0-2.0*mMatt) * (StarState['Rstar']*const.Rsun)**(4.0*mMatt+2.0) * StarState['OmegaEnv'] * const.OmegaSun / ( K2Matt**2.0 * StarState['vEsc']**2.0 + (StarState['OmegaEnv']*const.OmegaSun)**2.0 * (StarState['Rstar']*const.Rsun)**2.0 )**mMatt
    
    # Multiply by extra factor and make negitive so it is a spin-down torque
    torqueWind = - params['Kwind'] * torqueWind
    
    return torqueWind

def _Omega(Prot):
    """Takes rotation period in days, returns angular velocity in OmegaSun."""
    return 2.0*const.Pi / ( Prot*const.day ) / const.OmegaSun

def _Prot(Omega):
    """Takes angular velocity in OmegaSun, returns rotation period in days."""
    return 2.0*const.Pi / ( Omega*const.OmegaSun ) / const.day

def _Ro(StarState):
    """Takes state of star, returns Rossby number."""
    
    # Get Rossby 
    Ro = StarState['Prot'] / StarState['tauConv']
    
    return Ro

def _Bdip(StarState,params=params.paramsDefault):
    """Takes state of star, returns dipole field strength."""
    
    # Use Rossby number to get dipole field, depending if saturated or not
    if ( StarState['Ro'] >= params['RoSatBdip'] ):
        BdipStar = params['BdipSun'] * ( StarState['Ro'] / const.RoSun )**params['aBdip']
    else:
        BdipStar = params['BdipSun'] * ( params['RoSatBdip'] / const.RoSun )**params['aBdip']
    
    return BdipStar

def _Mdot(StarState,params=params.paramsDefault):
    """Takes state of star, returns mass loss rate."""

    # Use Rossby number to get mass loss rate
    if ( StarState['Ro'] >= params['RoSatMdot'] ):
        MdotWind = params['MdotSun'] * StarState['Rstar']**2.0 * (StarState['Ro']/const.RoSun)**params['aMdot'] * StarState['Mstar']**params['bMdot']
    else:
        MdotWind = params['MdotSun'] * StarState['Rstar']**2.0 * (params['RoSatMdot']/const.RoSun)**params['aMdot'] * StarState['Mstar']**params['bMdot']
    
    # Multiply Mdot by additional mutliplicative factor for rapid rotators if desired
    if params['BreakupMdotIncrease']:
        MdotWind *= MdotFactor(StarState['Mstar'],StarState['Rstar'],StarState['OmegaEnv'])
    
    return MdotWind

def MdotFactor(Mstar,Rstar,OmegaSurface,params=params.paramsDefault):
    """Takes stellar mass, radius, and rotation rate, returns factor to increase wind mass loss rate."""
    
    # Initially assume factor is unity
    factor = 1.0
    
    # Rotation fraction of breakup
    fracBreak = OmegaSurface / OmegaBreak(Mstar,Rstar)
    
    # Work out if close to break up 
    if ( fracBreak > params['fracBreakThreshold'] ):
        
        # The equation to use should depend on the fraction of break-up that the star is at
        # basically I (C.P.Johnstone) did two fits and the better fit is the first one here
        # but this is unstable when fracBreak>1.009 so at that point, the code switches to 
        # the other fit. This should not actually happen since stars should not get that 
        # rapidly rotating. 
        if ( fracBreak < 1.009 ):
            
            # For the relation below, do not allow frac larger than 1.009 since that
            # gives us NaNs 
            fracBreak = min( fracBreak , 1.009 )
            
            # Get the factor using a better fit to the factor-frac relation based on Johnstone (2017)
            factor = 0.93 * ( 1.01 - fracBreak )**(-0.43) * np.exp( 0.31 * fracBreak**7.5 )
            
        else:
            
            #Get the multiplicative factor (ORIGINAL NOT SO GOOD EQUATION)
            factor = (5.94401988e+03) * fracBreak**7.0 + (-2.12411847e+04) * fracBreak**6.0 + (3.07933216e+04) * fracBreak**5.0 + (-2.32693067e+04) * fracBreak**4.0 + (9.80417648e+03) * fracBreak**3.0 + (-2.27923380e+03) * fracBreak**2.0 + (2.68629816e+02) * fracBreak**1.0 + (-1.13054867e+01) * fracBreak**0.0
        
    return factor

def OmegaBreak(Mstar,Rstar):
    """Takes stellar mass and radius in solar units, returns breakup rotation rate in OmegaSun.
        
    Parameters
    ----------
    Mstar : float 
        Mass of star in Msun.
    Rstar : float 
        Radius of star in Rsun.
    
    Returns
    ----------
    OmegaBreak : float
        Breakup angular velocity in OmegaSun.
    
    """
    
    # Assuming here that Rpole = Rstar  
    return (2.0/3.0)**(3.0/2.0) * ( const.GravConstant * Mstar*const.Msun )**0.5 / ( Rstar*const.Rsun )**(3.0/2.0) / const.OmegaSun

def _vEsc(Mstar,Rstar):
    """Takes stellar mass and radius in solar units, returns surface escape velocity in cm s^-1.
        
    Parameters
    ----------
    Mstar : float 
        Mass of star in Msun.
    Rstar : float 
        Radius of star in Rsun.
    
    Returns
    ----------
    vEsc : float
        Surface escape velocity in cm s^-1.
    
    """
    
    return ( 2.0 * const.GravConstant * Mstar*const.Msun / (Rstar*const.Rsun) )**0.5

def _shouldCoreEnvelopeDecoupling(StarState,params=params.paramsDefault):
    """Takes state of star, determines if core-envelope decoupling should be used."""
    
    # For this, there are several things that need to be fulfilled:-
    #     1. The parameter CoreEnvelopeDecoupling must be set to True
    #     2. The ratio of the core to total stellar moment of inertia must be above the parameter IcoreThresholdCE
    #     3. The stellar mass must be above the parameter MstarThresholdCE
    
    # Test condition 1
    if not ( params['CoreEnvelopeDecoupling'] ):
        return False
    
    # Test condition 2
    if ( StarState['Icore']/StarState['Itotal'] < params['IcoreThresholdCE'] ):
        return False
    
    # Test condition 3
    if ( StarState['Mstar'] < params['MstarThresholdCE'] ):
        return False
    
    # If got to here, then core-envelope decoupling should happen
    return True

def OmegaSat(Mstar=None,Age=None,param='XUV',params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns saturation rotation rate as surface angular velocity.
    
    The user can supply a mass and age, and the code will return the rotation rate of the saturation threshold
    as a multiple of the solar rotation rate (2.67e-6 rad s^-1). The user can optionally also give a parameter 
    dictionary and an instance of the StarEvo class, though  this is not necessary and if these are not specified 
    then the defaults will be used. The code has three activity quantities (XUV, Mdot, and Bdip) that saturate
    and by default it is assumed that they saturate at the same rotation rates, but there are three separate
    model parameters for each (RoSatBdip, RoSatMdot, RoSatXray). By default, it is assumed that the user wants 
    the X-ray saturation threshold but this can be changed with the param keyword argument.
    
    Parameters
    ----------
    Mstar : float or np.ndarray
        Mass of star in Msun.
    Age : float or np.ndarray
        Age of star in Myr.
    param : str , optional
        String specifying which quantity the saturation threshold is needed for (default='XUV', options: 'XUV', 'Bdip', 'Mdot').
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    Omega : float
        Saturation rotation rate in OmegaSun.
    
    """
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Get rotation period of saturation threshold
    Prot = ProtSat(Mstar=Mstar,Age=Age,param=param,params=params,StarEvo=StarEvo)
    
    # Convert to rotation angular velocity in OmegaSun
    Omega = _Omega(Prot)
    
    return Omega

def ProtSat(Mstar=None,Age=None,param='XUV',params=params.paramsDefault,StarEvo=None):
    """Takes basic stellar parameters, returns saturation rotation rate as rotation period.
    
    The user can supply a mass and age, and the code will return the rotation rate of the saturation threshold
    as a multiple of the solar rotation rate (2.67e-6 rad s^-1). The user can optionally also give a parameter 
    dictionary and an instance of the StarEvo class, though  this is not necessary and if these are not specified 
    then the defaults will be used. The code has three activity quantities (XUV, Mdot, and Bdip) that saturate
    and by default it is assumed that they saturate at the same rotation rates, but there are three separate
    model parameters for each (RoSatBdip, RoSatMdot, RoSatXray). By default, it is assumed that the user wants 
    the X-ray saturation threshold but this can be changed with the param keyword argument.
    
    Parameters
    ----------
    Mstar : float
        Mass of star in Msun.
    Age : float
        Age of star in Myr.
    param : str , optional
        String specifying which quantity the saturation threshold is needed for (default='XUV', options: 'XUV', 'Bdip', 'Mdot').
    params : dict , optional
        Dictionary holding model parameters.
    StarEvo : mors.stellarevo.StarEvo , optional
        Instance of StarEvo class holding stellar evolution model data.
    
    Returns
    ----------
    Prot : float
        Saturation rotation period in days.
    
    """
    
    # If an instance of the StarEvo class is not input, load it with defaults
    if StarEvo is None:
        StarEvo = SE.StarEvo()
    
    # Get saturation Rossby number based on user input of param
    if ( param == 'XUV' ):
        Ro = params['RoSatXray']
    elif ( param == 'Bdip' ):
        Ro = params['RoSatBdip']
    elif ( param == 'Mdot' ):
        Ro = params['RoSatMdot']
    else:
        misc._PrintErrorKill("invalid value of param in call to function")
    
    # Convective turnover time in days
    tauConv = StarEvo.tauConv(Mstar,Age)
    
    # Work out rotation period saturation
    Prot = Ro * tauConv
    
    return Prot

def aOrbHZ(Mstar=None,Age=None,params=params.paramsDefault):
    """Takes stellar mass, returns orbital distances of habitable zone boundaries.
    
    This function can be used to get the habitable zone boundary orbital distances as a function of stellar mass
    and age. The mass can be either a simple value or an array, in which case all HZ boundaries will be returned
    as arrays. The age does not need to be set and by default the value of the AgeHZ parameter (default 5000 Myr)
    will be used. The habitable zone boundaries are calculated using the formulae of Kopparapu et al. (2013) and
    the luminosities and effective temperatures from the stellar models of Spada et al. (2013). The dictionary that
    is returned contains the following elements: 'RecentVenus', 'RunawayGreenhouse', 'MoistGreenhouse', 
    'MaximumGreenhouse', 'EarlyMars', and 'HZ'. While the first five are obvious and correspond to the boundaries 
    in Kopparapu et al. (2013), the final one is defined in Johnstone et al. (2020) as the average of the runaway 
    greenhouse and moist greenhouse orbital distances. All are calculated in AU.
    
    Parameters
    ----------
    Mstar : float or np.ndarray
        Stellar mass in Msun.
    Age : float , optional
        Age to get stellar parameters in Myr (default = 5000 Myr).
    params : dict , optional
        Dictionary holding model parameters.
    Returns
    ----------
    aOrbHZAll : dict
        Values of HZ boundary orbital distances in AU.
    """
    
    # Make sure Mstar is set 
    if Mstar is None:
        misc._PrintErrorKill("Mstar must be set in call to function")
    
    # If Age is not set, get value from params
    if Age is None:
        Age = params['AgeHZ']
    
    # Start dictionary
    aOrbHZAll = {}
    
    # Set effective temperatures and luminosities of stars
    Lbol = SE.Lbol( Mstar , Age )
    Teff = SE.Teff( Mstar , Age )
    
    # Converf Teff to Tstar
    Teff += -5780.0
    
    # Get Recent Venus limit
    SeffSun = 1.7763
    a = 1.4335e-4
    b = 3.3954e-9
    c = -7.6364e-12
    d = -1.1950e-15
    Seff = SeffSun + a*Teff + b*Teff**2.0 + c*Teff*3.0 + d*Teff**4.0
    aOrbHZAll['RecentVenus'] = ( Lbol / Seff )**0.5
    
    # Get Runaway Greenhouse limit
    SeffSun = 1.0385
    a = 1.2456e-4
    b = 1.4612e-8
    c = -7.6345e-12
    d = -1.7511e-15
    Seff = SeffSun + a*Teff + b*Teff**2.0 + c*Teff*3.0 + d*Teff**4.0
    aOrbHZAll['RunawayGreenhouse'] = ( Lbol / Seff )**0.5
    
    # Get Moist Greenhouse limit
    SeffSun = 1.0146
    a = 8.1884e-5
    b = 1.9394e-9
    c = -4.3618e-12
    d = -6.8260e-16
    Seff = SeffSun + a*Teff + b*Teff**2.0 + c*Teff*3.0 + d*Teff**4.0
    aOrbHZAll['MoistGreenhouse'] = ( Lbol / Seff )**0.5
    
    # Get Maximum Greenhouse limit
    SeffSun = 0.3507
    a = 5.9578e-5
    b = 1.6707e-9
    c = -3.0058e-12
    d = -5.1925e-16
    Seff = SeffSun + a*Teff + b*Teff**2.0 + c*Teff*3.0 + d*Teff**4.0
    aOrbHZAll['MaximumGreenhouse'] = ( Lbol / Seff )**0.5
    
    # Get Early Mars limit
    SeffSun = 0.3207
    a = 5.4471e-5
    b = 1.5275e-9
    c = -2.1709e-12
    d = -3.8282e-16
    Seff = SeffSun + a*Teff + b*Teff**2.0 + c*Teff*3.0 + d*Teff**4.0
    aOrbHZAll['EarlyMars'] = ( Lbol / Seff )**0.5
    
    # Now get value of center of moist and maximum greenhouse
    aOrbHZAll['HZ'] = 0.5 * ( aOrbHZAll['MoistGreenhouse'] + aOrbHZAll['MaximumGreenhouse'] )
    
    return aOrbHZAll



"""Module for historical spectral synthesis"""

# Import system libraries 
import numpy as np
from copy import deepcopy
from scipy.optimize import minimize

# Import MORS files 
import mors.spectrum as spec
import mors.constants as const
import mors.miscellaneous as misc
from mors.star import  Percentile
from mors.stellarevo import Value, Lbol
from mors.physicalmodel import Lxuv


def GetProperties(Mstar:float, pctle:float, age:float):
    """Calculate properties of star for a given rotation percentile and  age
    
    Parameters 
    ----------
        Mstar : float 
            Mass of star [M_sun]
        pctle : float 
            Rotation percentile 
        age : float 
            Stellar age  [Myr] 

    Returns 
    ----------
        out : dict
            Dictionary of radius [m], Teff [K], and band fluxes at 1 AU [erg s-1 cm-2]
    """

    # Get star radius [m]
    Rstar = Value(Mstar, age, 'Rstar') * const.Rsun * 1.0e-2

    # Get star temperature [K]
    Tstar = Value(Mstar, age, 'Teff')

    # Get rotation rate [Omega_sun]
    Omega = Percentile(Mstar=Mstar, percentile=pctle)

    # Get luminosities and fluxes 
    Ldict = Lxuv(Mstar=Mstar, Age=age, Omega=Omega)

    # Output 
    out = {
        "mass"   : Mstar,      # units of M_sun 
        "pctle"  : pctle,
        "age"    : age,        # units of Myr
        "radius" : Rstar,
        "Teff"   : Tstar,
    }

    # Luminosities (erg s-1)
    out["L_bo"] = Lbol(Mstar,age) * const.LbolSun   
    out["L_xr"] = Ldict["Lx"] 
    out["L_e1"] = Ldict["Leuv1"] 
    out["L_e2"] = Ldict["Leuv2"] 

    # Fluxes at 1 AU
    area = (4.0 * const.Pi * const.AU * const.AU)
    for k in ["bo","xr","e1","e2"]:
        out["F_"+k] = out["L_"+k]/area

    # Get flux from Planckian band 
    wl_pl = np.logspace(np.log10(spec.bands_limits["pl"][0]), np.log10(spec.bands_limits["pl"][1]), 1000)
    fl_pl = spec.PlanckFunction_surf(wl_pl, Tstar)
    fl_pl = spec.ScaleTo1AU(fl_pl, Rstar)
    out["F_pl"] = np.trapz(fl_pl, wl_pl)
    out["L_pl"] = out["F_pl"] * area

    # Get flux of UV band from remainder 
    out["F_uv"] = out["F_bo"] - out["F_xr"] - out["F_e1"] - out["F_e2"] - out["F_pl"]
    out["L_uv"] = out["F_uv"] * area

    return out 


def CalcBandScales(modern_dict:dict, historical_age:float):
    """Get band scale factors for historical spectrum

    Parameters 
    ----------
        modern_dict : dict 
            Dictionary output of `GetProperties` call for modern spectrum
        historical_age : float 
            Stellar age at which to get corresponding scale factors

    Returns 
    ----------
        Q_dict : dict
            Dictionary of band scale factors
    """

    # Get properties 
    historical = GetProperties(modern_dict["mass"], modern_dict["pctle"], historical_age)

    # Get scale factors 
    Q_dict = {}
    for key in spec.bands_limits.keys():
        Q_dict["Q_"+key] = historical["F_"+key]/modern_dict["F_"+key]

    return Q_dict

def CalcScaledSpectrumFromProps(modern_spec:spec.Spectrum, modern_dict:dict, historical_age:float):
    """Scale a stellar spectrum according to stellar properties.

    Returns a new Spectrum object containing the historical fluxes.

    Parameters 
    ----------
        modern_spec : Spectrum object
            Spectrum object containing data for a modern fluxes
        modern_dict : dict 
            Dictionary output of `GetProperties` call for modern spectrum
        historical_age : float 
            Stellar age at which to get corresponding scale factors

    Returns 
    ----------
        historical_spec : Spectrum object 
            Spectrum object containing data for historical fluxes
    """

    # Get scale factors relative to modern spectrum
    Q_dict = CalcBandScales(modern_dict, historical_age)

    # Get modern wl, fl
    spec_fl = deepcopy(modern_spec.fl)
    spec_wl = deepcopy(modern_spec.wl)

    # Get band indicies 
    for i in range(len(spec_wl)):
        b = spec.WhichBand(spec_wl[i])
        if b == None:
            continue 
        spec_fl[i] *= Q_dict["Q_"+b[0]]

    # Make new spectrum object
    historical_spec = spec.Spectrum()
    historical_spec.LoadDirectly(spec_wl, spec_fl)

    return historical_spec

    
def FitModernProperties(modern_spec:spec.Spectrum, Mstar:float, age:float=-1):
    """Estimate rotation percentile and (optionally) age from modern spectrum.

    Parameters 
    ----------
        modern_spec : Spectrum object
            Spectrum object containing data for a modern fluxes
        Mstar : float 
            Stellar mass [M_sun]
        age : float 
            Optional guess for current age. Will be estimated if not provided.

    Returns 
    ----------
        best_pctle : float 
            Best estimate of rotation percentile 
        best_age : float 
            Best estimate of star's age (equal to `age` if `age` is provided)
    """

    # Integrated fluxes
    modern_spec.CalcBandFluxes()

    fit_age = bool(age>0)

    # Objective function
    def _fev(x_arr:tuple):

        this_pctle = x_arr[0]
        if fit_age:
            this_age = x_arr[1]
        else:
            this_age = age

        this_age = max(min(this_age, 11000), 0.4)

        props = GetProperties(Mstar, this_pctle, this_age)

        resid = 0.0
        for k in ["xr","e1","e2","uv"]:
            lim = spec.bands_limits[k]
            wid = lim[1]-lim[0]
            resid += (props["F_"+k]/wid - modern_spec.fl_integ[k]/wid)**2
        
        return np.sqrt(resid)
    
    # Initial guess 
    if fit_age:
        x0 = [1.0, 1000.0]
    else:
        x0 = [1.0]

    # Find best params
    result = minimize(_fev, x0, method='Nelder-Mead')

    # Check 
    if not result.success:
        misc._PrintError("Could not fit stellar properties to modern spectrum")

    # Result 
    best_pctle = result.x[0]
    if fit_age:
        best_age = result.x[1]
    else:
        best_age = age

    # Return
    return best_pctle, best_age
    





"""Module holding the Star class and related functions."""

# Imports for standard stuff needed here
import sys
import inspect
import numpy as np
import pickle

# Imports for nors modules
import mors.miscellaneous as misc
import mors.stellarevo as SE
import mors.parameters as params
import mors.rotevo as RE
import mors.physicalmodel as phys

# Limits on various input values
MstarMin = 0.1
MstarMax = 1.25

class Star:
    """A class for star objects that hold all information about a star. 
    
    Attributes
    ------------
    
    Methods
    ------------
    
    """
    
    #---------------------------------------------------------------------------------------

    def __init__(self,Mstar=None,Age=None,percentile=None,Omega=None,Prot=None,OmegaEnv=None,OmegaCore=None,AgesOut=None,starEvoDir=None,evoModels=None,params=params.paramsDefault):
        """Initialises instance of Star class.
        
        This is the main function that is run when creating an instance of the Star class and it sets up all
        the things needed including calculating evolutionary tracks for the star. The function requires that
        the arguments Mstar (the star's mass in Msun) and Omega (in OmegaSun=2.67e-6 rad s^-1) are specified
        in the call. Alternatively, OmegaEnv and OmegaCore can be set, in which case Omega does not need to 
        be specified. If the argument Age (in Myr) is also specified, then the code will find the evolutionary
        track that passes through this rotation rate at this age, otherwise if Age is not set then it will 
        calculate evolutionary tracks assuming this Omega as the initial (1 Myr) rotation rate. The user should
        not specify OmegaCore and Age simultaneously, and if Age is set then either Omega or OmegaEnv can be
        used to specify the surface rotation rate. The user can also specify the initial rotation rate using 
        
        Parameters
        ----------
        Mstar : float
            Stellar mass in Msun.
        Age : float , optional
            Age in Myr of the input rotation rate (default = 1 Myr).
        Omega : float , optional
            Rotation rate of star at Age in OmegaSun.
        Prot : float , optional
            Rotation period of star at Age in days.
        OmegaEnv : float , optional
            Envelope rotation rate of star at Age in OmegaSun.
        OmegaCore : float , optional
            Core rotation rate of star at Age in OmegaSun.
        percentile : float , optional
            Percentile in 1 Myr distribution (between 0 and 100).
            
        params : dict , optional
            Parameter dictionary used for getting width of bin to use for distribution.
        
        Returns
        ----------
        None
            None
        
        """
        
        # Make sure Mstar has good values
        _CheckInputMstar(Mstar)
        
        # Make sure Omega, OmegaEnv, and OmegaCore are well set
        # (if Omega is set, OmegaEnv and OmegaCore will both be set to that value)
        Omega , OmegaEnv , OmegaCore = _InputRotation(Mstar,Age,Omega,OmegaEnv,OmegaCore,Prot,percentile,params)
        
        # Set parameters
        self.params = params
        
        # Set the ExtendedTracks parameter to True so we get all parameters
        self.params['ExtendedTracks'] = True
        
        # Set Mstar
        self.Mstar = Mstar
        
        # Set stellar evo model
        self.starEvoDir = starEvoDir
        self.evoModels = evoModels
        
        # Load evo tracks if Mstar is set
        self._LoadStarEvo()
        
        # Get evolutionary tracks 
        self._LoadEvoTracks(Age,OmegaEnv,OmegaCore,AgesOut=AgesOut)
        
        # Get HZ boundaries
        self.aOrbHZ = phys.aOrbHZ(Mstar=self.Mstar,params=self.params)
        
        return
    
    def _LoadStarEvo(self,starEvoDir=None,evoModels=None):
        """Loads stellar evolution tracks."""
        
        # Get an instance of the StarEvo class holding evolutionary tracks for all masses
        self.StarEvo = SE.StarEvo(starEvoDir=self.starEvoDir,evoModels=self.evoModels)
        
        # Load evolutionary tracks for this mass only (so delete tracks for other masses after getting this one)
        self.StarEvo.LoadTrack(self.Mstar,ClearData=True)
        
        return
    
    def _LoadEvoTracks(self,Age,OmegaEnv0,OmegaCore0,AgesOut=None):
        """Loads rotation and activity tracks."""
        
        
        # Set starting age to 1 Myr
        self.AgeMin = 1.0
        
        # Get ending age as end of stellar evolution track
        self.AgeMax = self.StarEvo.ModelData[self.Mstar]['Age'][-1]
        
        # If needing to get initial rotation rate then get it
        if not Age is None:
            
            # Get initial rotation
            Omega0 = RE.FitRotation( Mstar=self.Mstar , Age=Age , Omega=OmegaEnv0 , 
                                        AgeMin=self.AgeMin , params=self.params , StarEvo=self.StarEvo )
            
            # Make sure a correct initial rotation rate was found
            if ( Omega0 == -1 ):
                misc._PrintErrorKill("input rotation rate too low for given mass and age")
            if ( Omega0 == -2 ):
                misc._PrintErrorKill("input rotation rate too high for given mass and age")
            if ( Omega0 == -3 ):
                misc._PrintErrorKill("input rotation rate in valid range for mass and age but solver was unable to fit track")
            
            # Set initial rotation rates for input into RE.EvolveRotation()
            OmegaEnv0 = Omega0
            OmegaCore0 = Omega0
            
        # Run evolution model using OmegaEnv0 and OmegaCore0 as initial rotation rate
        self.Tracks = RE.EvolveRotation( Mstar=self.Mstar , OmegaEnv0=OmegaEnv0 , OmegaCore0=OmegaCore0 , 
                                        AgeMin=self.AgeMin , AgeMax=self.AgeMax , AgesOut=AgesOut ,
                                        params=self.params , StarEvo=self.StarEvo )
        
        # Set the percentile of this track in the starting distribution
        self.percentile = Percentile( Mstar=self.Mstar , Omega=OmegaEnv0 )
        
        # Make evolutionary tracks attributes of this class 
        # (this is so the user can get a track using e.g. star.Age instead of star.Tracks['Age'])
        for track in self.Tracks:
            
            # Note: the following two options are basically equivalent
            #self.__dict__[track+'Track'] = self.Tracks[track]
            setattr( self , track+'Track' , self.Tracks[track] )
        
        # Load dictionary holding units for each quantity
        self.Units = phys.QuantitiesUnits()
        
        # Make functions for each quantity that return this quantity at a given age as attributes of class
        self._setupQuantityFunctions()
        
        return
            
    def _setupQuantityFunctions(self):
        """Makes functions for each quantity that return this quantity at a given age as attributes of class."""
        
        # Loop over quantities held in self.Tracks
        for track in self.Tracks:
            
            # The method for this is very strange and I don't like it. Better would likely just to be to 
            # add individual functions for each quantity since I was unable to find a solution that did 
            # not involve using exec() here. This should not provide a security risk since exec() is not
            # being run on any user input data. The following two lines first create a two-line function
            # from a string that calls self.Value() with a specific value for Quantity, so the function
            # Lx() that is created calls Value with Quantity='Lx'. This function is then added to the Star 
            # class using setattr on self.__class__. Technically this means that if the user makes multiple
            # instances of Star, the next two lines will not be necessary.
            exec( "def "+track+"(self,Age):\n  return self.Value(Age=Age,Quantity='"+track+"')" )
            exec( "setattr( self.__class__ , '"+track+"' , "+track+" )" )
            
        return
        
    def Value(self,Age=None,Quantity=None):
        """Takes age in Myr and a string with name of quantity to output, returns value of that quantity at specified age."""
        
        # Do 1D linear interpolation to get value
        value = SE._Interpolate1D( self.Tracks['Age'] , self.Tracks[Quantity] , Age )
        
        return value
    
    def Track(self,Quantity=None):
        """Takes a quantity string, return evolutionary track for that quantity."""
        
        # Make sure Quantity is set
        if Quantity is None:
            misc._PrintErrorKill("Quantity not set in call to function")
        
        # Make sure Quantity is string
        if not isinstance(Quantity,str):
            misc._PrintErrorKill("Quantity must be string")
        
        # Make sure Quantity is a valid quantity with a track
        if not Quantity in self.Tracks:
            misc._PrintErrorKill("no available track for "+Quantity)
        
        return self.Tracks[Quantity]
  
    def PrintAvailableTracks(self):
        """Prints all quantities that have evolutionary tracks available."""
        
        # Print header
        print("")
        print("EVOLUTIONARY TRACKS AVAILABLE FOR:-")
        
        # Loop over each element in Tracks and print them but only if they are numpy arrays
        for track in self.Tracks:
            if ( type(self.Tracks[track]).__module__ == 'numpy' ):
                
                # Start string with small space and name of quantity
                outString = "   "+track
                
                # If the units are given, add that in parentheses
                if not ( self.Units[track] == '' ):
                    outString += " ("+self.Units[track]+")"
                
                # Print the string 
                print(outString)
            
        # A free line at the end to look better
        print("")
        
        return
   
    def Save(self,filename='star.pickle'):
        """Takes filename (default is 'star.pickle'), saves cluster to this file using pickle."""
        with open(filename,'wb') as f:
            pickle.dump(self,f)
        return
  
    def save(self,filename='star.pickle'):
        """Same as Save()."""
        self.Save(filename=filename)
        return
  
    def ActivityLifetime(self,Quantity=None,Threshold=None,AgeMax=None):
        """Takes threshold value, returns age at which a star last drops below this threshold.
        
        This function can be used to determine when a star's emission crosses a given threshold value for a few
        activity quantities. These are Lx, Fx, Rx, and FxHZ for X-rays, and similar values for EUV1, EUV2, EUV, 
        XUV, and Ly-alpha. If the star crosses the threshold (from above it to below it) multiple times, this 
        will find the final time it will cross the threshold. If the user wants to set a maximum age so that the 
        code only looks for crossings of the threshold below this age then this can be done using the AgeMax
        keyword argument.
            
        Parameters
        ----------
        Quantity : str
            Gives which quantity to consider (e.g. 'Lx').
        Threshold : float or str
            Value for threshold in units of quantity or string of 'sat'.
        AgeMax : float , optional
            End age of track to consider in Myr.
        
        Returns
        ----------
        AgeActive : float
            Activity lifetime in Myr.
        
        """
        
        # Allowed quantities
        QuantitiesAllowed = [ 'Lx' , 'Fx' , 'Rx' , 'FxHZ' , 'Leuv1' , 'Feuv1' , 'Reuv1' , 'Feuv1HZ' , 'Leuv2' , 'Feuv2' , 'Reuv2' , 'Feuv2HZ' , 
                            'Leuv' , 'Feuv' , 'Reuv' , 'FeuvHZ' , 'Lly' , 'Fly' , 'Rly' , 'FlyHZ' ]
        
        # Make sure Quantity is set
        if Quantity is None:
            misc._PrintErrorKill("Quantity not set in call to function")
        
        # Make sure Quantity is string
        if not isinstance(Quantity,str):
            misc._PrintErrorKill("Quantity must be string")
            
        # Check input Quantity is valid
        if not Quantity in QuantitiesAllowed:
            QuantitiesString = ''
            for quantity in QuantitiesAllowed:
                QuantitiesString += quantity + " , "
            misc._PrintErrorKill("invalid quantity "+Quantity+"\n valid options are "+QuantitiesString[0:-3])
        
        # Set tracks
        AgeTrack = self.AgeTrack
        Track = self.Tracks[Quantity]
            
        # If Threshold is 'sat' then normalise track to saturation value and set threshold to unity
        if ( Threshold == 'sat' ):
            
            # Get saturation rotation rate as function of age
            OmegaSatTrack = phys.OmegaSat( Mstar=self.Mstar , Age=self.AgeTrack , param='XUV' , params=self.params , StarEvo=self.StarEvo )
            
            # Make track just OmegaEnv/OmegaSat
            Track = self.OmegaEnvTrack / OmegaSatTrack
            
            # Set threshold to unity
            Threshold = 1.0
        
        # Get result
        AgeActive = misc.ActivityLifetime( Age=AgeTrack , Track=Track , Threshold=Threshold , AgeMax=AgeMax )
        
        return AgeActive
 
    def IntegrateEmission(self,AgeMin=None,AgeMax=None,Band=None,aOrb=None):
        """Takes age range, calculates integrated emission in band within that range.
        
        This code can be used to integrate a star's luminosity between two ages. This can be applied to any wavelength band
        and the result is a total energy emitted in this time in erg. If the user also specifies an orbital distance using 
        the aOrb keyword argument, the code integrates the flux at this obital distance returns a fluence in erg cm^-2. The
        user can specify aOrb as a string to get the fluences at various habitable zone boundaries (using the HZ calculated 
        at the age defined in params when creating this cluster). Options are 'RecentVenus', 'RunawayGreenhouse', 'MoistGreenhouse', 
        'MaximumGreenhouse', 'EarlyMars', and 'HZ'.
            
        Parameters
        ----------
        AgeMin : float
            Start of time period to integrate in Myr.
        AgeMax : float
            End of time period to integrate in Myr.
        Band : str
            Gives which wavelength band to consider (options are 'XUV', 'Xray', 'EUV1', 'EUV2', 'EUV', 'Lyman', 'bol').
        aOrb : float or str , optional
            Orbital distance to get fluence at in AU or string identifying HZ boundary.
        
        Returns
        ----------
        Energy : float
            Integrated luminosity or flux in erg or erg cm^-2.
        
        """
        
        # Allowed quantities
        BandsAllowed = [ 'XUV', 'Xray', 'EUV1', 'EUV2', 'EUV', 'Lyman', 'bol' ]
        
        # Make sure Band is set
        if Band is None:
            misc._PrintErrorKill("Band not set in call to function")
        
        # Make sure Band is string
        if not isinstance(Band,str):
            misc._PrintErrorKill("Band must be string")
            
        # Check input Band is valid
        if not Band in BandsAllowed:
            BandsString = ''
            for band in BandsAllowed:
                BandsString += band + " , "
            misc._PrintErrorKill("invalid Band "+Band+"\n valid options are "+BandsString[0:-3])
            
        # Get luminosity track to use
        if Band == 'XUV':
            Luminosity = self.LxTrack + self.LeuvTrack
        elif Band == 'Xray':
            Luminosity = self.LxTrack
        elif Band == 'EUV1':
            Luminosity = self.Leuv1Track
        elif Band == 'EUV2':
            Luminosity = self.Leuv2Track
        elif Band == 'EUV':
            Luminosity = self.LeuvTrack
        elif Band == 'Lyman':
            Luminosity = self.LlyTrack
        elif Band == 'bol':
            Luminosity = self.LbolTrack
        else:
            misc._PrintErrorKill("did not find right luminosity track")
            
        # If aOrb was set to a string, do necessary work to get useable aOrb (given by aOrbUse)
        if isinstance(aOrb,str):
            
            # Valid options
            aOrbAllowed = [ 'RecentVenus' , 'RunawayGreenhouse' , 'MoistGreenhouse' , 'MaximumGreenhouse' , 'EarlyMars' , 'HZ' ]
            
            # Check input is valid
            if not aOrb in aOrbAllowed:
                aOrbString = ''
                for a in aOrbAllowed:
                    aOrbString += a + " , "
                misc._PrintErrorKill("invalid aOrb "+aOrb+"\n valid options are a value in AU or "+aOrbString[0:-3])
            
            # Get orbital distance
            aOrbUse = self.aOrbHZ[aOrb]
            
        else:
            
            # Just set to input value
            aOrbUse = aOrb
            
        # Do calculation
        Energy = misc.IntegrateEmission( AgeMin=AgeMin , AgeMax=AgeMax , Age=self.AgeTrack , Luminosity=Luminosity , aOrb=aOrbUse )
        
        return Energy

def _CheckInputMstar(Mstar):
    """Takes stellar mass, checks if value is a valid input mass."""
    
    # Make sure Mstar was set
    if Mstar is None:
        misc._PrintErrorKill("stellar mass not given")
    
    # Make sure it is a float
    #if not isinstance(Mstar,float):
        #misc._PrintErrorKill("stellar mass must be given as float")
    
    # Make sure it above minimum
    if ( Mstar < MstarMin ):
        misc._PrintErrorKill( "stellar mass cannot be less than lower limit of "+str(MstarMin)+" Msun" )
    
    # Make sure it above minimum
    if ( Mstar > MstarMax ):
        misc._PrintErrorKill( "stellar mass cannot be greater than upper limit of "+str(MstarMax)+" Msun" )
        
    return

def _InputRotation(Mstar,Age,Omega,OmegaEnv,OmegaCore,Prot,percentile,params):
    """Takes input rotation values, checks sets up values correctly."""
    
    # Make sure percentile and rotation are not both set
    if not percentile is None: 
        if not Omega is None :
            misc._PrintErrorKill( "cannot set both percentile and Omega as arguments of Star" )
        if not OmegaEnv is None :
            misc._PrintErrorKill( "cannot set both percentile and OmegaEnv as arguments of Star" )
        if not OmegaCore is None :
            misc._PrintErrorKill( "cannot set both percentile and OmegaCore as arguments of Star" )
        if not Prot is None :
            misc._PrintErrorKill( "cannot set both percentile and Prot as arguments of Star" )
        
    # Make sure percentile and OmegaEnv are not both set
    if ( not percentile is None ) and ( not OmegaEnv is None ):
        misc._PrintErrorKill( "cannot set both percentile and OmegaEnv as arguments of Star" )
    
    # If percentile was set to a string, get float version
    if isinstance(percentile,str):
        if ( percentile == 'slow' ):
            percentile = 5.0
        elif ( percentile == 'medium' ):
            percentile = 50.0
        elif ( percentile == 'fast' ):
            percentile = 95.0
        else:
            misc._PrintErrorKill( "invalid percentile string (options are 'slow', 'medium', or 'fast')" )
        
    # If percentile was set, get Omega
    if not percentile is None:
        Omega = Percentile(Mstar=Mstar,percentile=percentile,params=params)
    
    # Make sure if Age is set that OmegaCore is not set
    if ( not Age is None ) and ( not OmegaCore is None ):
        misc._PrintErrorKill( "cannot set both Age and OmegaCore as arguments of Star" )
    
    # Make sure not both Omega and Prot were set
    if ( not Omega is None ) and ( not Prot is None ):
        misc._PrintErrorKill( "cannot set both Omega and Prot as arguments of Star" )
    
    # If Prot is set, get Omega
    if not Prot is None:
        Omega = phys._Omega(Prot)
    
    # Make sure at least one rotation rate is set
    if ( Omega is None ):
        
        # If Age is set then make sure OmegaEnv is set, otherwise make sure both OmegaEnv and OmegaCore are set
        if ( not Age is None ): 
            if ( OmegaEnv is None ):
                misc._PrintErrorKill( "if Age is set then must set either Omega or OmegaEnv as argument of Star" )
        elif ( OmegaEnv is None ) or ( OmegaCore is None ):
            misc._PrintErrorKill( "must set either Omega or both OmegaEnv and OmegaCore as argument of Star" )
    
    else:
        
        # Since Omega is set, make sure both OmegaEnv and OmegaCore are not set
        if not ( ( OmegaEnv is None ) and ( OmegaCore is None ) ):
            misc._PrintErrorKill( "cannot set OmegaEnv and OmegaCore when Omega is set as arugment of Star" )
        
        # Set both OmegaEnv and OmegaCore to Omega
        OmegaEnv = Omega
        OmegaCore = Omega
    
    return Omega , OmegaEnv , OmegaCore

def Percentile(Mstar=None,Omega=None,Prot=None,percentile=None,MstarDist=None,OmegaDist=None,ProtDist=None,params=params.paramsDefault):
    """Gets rotation rate of percentile or percentile of rotation rate in the model rotation distribution.
    
    This function can be used for two purposes
        1. to determine the percentile in a rotation distribution of a star given its mass and rotation rate
        2. to determine the rotation rate of a star in a rotation distribution given its mass and percentile
    In the first case, the user should specify the rotation rate using either the Omega or Prot keyword arguments (given 
    in OmegaSun and days respectively) and the mass. In the second case, the user should specify the percentile using the
    percentile keyword argument (in the range 0 to 100) and the mass. The mass should be specified in Msun using the Mstar
    keyword argument. These should only be given as floats. The user can also specify the rotation distribution using the 
    MstarDist and the OmegaDist or ProtDist keyword arguments. If this is not done, then the code will assume the 1 Myr 
    rotation distribution from the empirical model distribution used in Johnstone et al. (2020).
    
    Parameters
    ----------
    Mstar : float
        Stellar mass in Msun.
    Omega : float , optional
        Rotation rate of star in OmegaSun.
    Prot : float , optional
        Rotation period of star in days.
    percentile : float , optional
        percentile in distribution (between 0 and 100).
    MstarDist : numpy.ndarray , optional
        Array of masses of stars in distribution.
    OmegaDist : numpy.ndarray , optional
        Array of rotation rates of stars in distribution in OmegaSun.
    ProtDist : numpy.ndarray , optional
        Array of rotation periods of stars in distribution in days.
    params : dict , optional
        Parameter dictionary used for getting width of bin to use for distribution.
    
    Returns
    ----------
    result : float
        Either rotation rate for percentile or percentile for rotation rate.
    
    """
    
    # Make sure Mstar was set
    if Mstar is None:
        misc._PrintErrorKill( "keyword argument Mstar must be set" )
    
    # If percentile is not set, make sure rotation rate is set
    if percentile is None:
        # Not both
        if not Omega is None:
            if not Prot is None:
                misc._PrintErrorKill( "keyword arguments Omega and Prot cannot both be set" )
        # At least one
        if ( Omega is None ) and ( Prot is None ):
            misc._PrintErrorKill( "must set either Omega, Prot, or percentile keyword arguments" )
        # Set Omega is not set then set it
        if Omega is None:
            Omega = phys._Omega(Prot)
            
    # Setup the distribution
    if MstarDist is None:
        MstarDist , OmegaDist = misc.ModelCluster()
    else:
        
        # Check that one of OmegaDist and ProtDist are set
        if ( OmegaDist is None ) and ( ProtDist is None ):
            misc._PrintErrorKill( "one of keyword arguments OmegaDist and ProtDist must be set" )
        
        # Check that OmegaDist and ProtDist are not both set
        if ( not OmegaDist is None ) and ( not ProtDist is None ):
            misc._PrintErrorKill( "keyword arguments OmegaDist and ProtDist cannot both be set" )
        
        # Setup OmegaDist if ProtDist was set
        if OmegaDist is None:
            OmegaDist = misc._Omega(ProtDist)
        
        # Make sure arrays are same length
        if not ( len(MstarDist) == len(OmegaDist) ):
            misc._PrintErrorKill( "MstarDist and OmegaDist (or ProtDist) must be same length" )
            
    # Work out which one to do
    if not percentile is None:
        
        # percentile was set, so get corresponding rotation rates
        result = _OmegaPercentile(Mstar,percentile,MstarDist,OmegaDist,params)
    
    else:
        
        # Rotation rate was set, so get corresponding percentile
        result = _PerPercentile(Mstar,Omega,MstarDist,OmegaDist,params)
        
    
    return result 

def _OmegaPercentile(Mstar,percentile,MstarDist,OmegaDist,params):
    """Gets rotation rate that corresponds to a given percentile of the rotation distribution."""
    
    ## If percentile was a string, get actual value
    
    # Get part of distribution to include
    includeStars = np.where( np.abs(MstarDist-Mstar) <= params['dMstarPer'] )
    
    # Need at least two stars
    if ( len(includeStars[0]) < 2 ):
        misc._PrintErrorKill( "need at least two stars in mass bin to get percentile" )
    
    # Get percentile Omega
    Omega = np.percentile( OmegaDist[includeStars] , percentile )
    
    return Omega

def _PerPercentile(Mstar,Omega,MstarDist,OmegaDist,params):
    """Gets percentile of the rotation distribution for a given mass and rotation rate."""
    
    # Get part of distribution to include
    includeStars = np.where( np.abs(MstarDist-Mstar) <= params['dMstarPer'] )
    
    # Need at least two stars
    if ( len(includeStars[0]) < 2 ):
        misc._PrintErrorKill( "need at least two stars in mass bin to get percentile" )
    
    # If Omega is less than minimum then make this as the 0th percentile
    if ( Omega <= np.min(OmegaDist[includeStars]) ):
        return 0.0
    
    # If Omega is above maximum then make this as the 100th percentile
    if ( Omega >= np.max(OmegaDist[includeStars]) ):
        return 100.0
    
    # Use bisector to invert numpy.percentile function
    
    # Initial limits
    perMin = 0.0
    perMax = 100.0
    
    # Omegas for limits
    OmegaMin = np.percentile( OmegaDist[includeStars] , perMin )
    OmegaMax = np.percentile( OmegaDist[includeStars] , perMax )
    
    # Tolerance and maximum number of iterations
    tol = 1.0e-5
    nIterMax = 10000
        
    # Start iterating
    found = False
    for iIter in range(nIterMax):
        
        # Get mid value and percentle
        perMid = 0.5 * ( perMin + perMax )
        OmegaMid = np.percentile( OmegaDist[includeStars] , perMid )
        
        # See if OmegaMid is equal to or close enough to answer
        if ( abs(OmegaMid/Omega-1.0) < tol ):
            return perMid
        
        # See which limit to change
        if ( OmegaMid > Omega ):
            perMax = perMid
            OmegaMax = OmegaMid
        else:
            perMin = perMid
            OmegaMin = OmegaMid
        
    # Make sure it was found
    if not found:
        misc._PrintErrorKill( "did not find percentile "+str(per)+" for Omega of "+str(Omega) )
    
    return per




# The basic Star class that people should be using
from .star import Star , Percentile

# The basic Cluster class that people should be using
from .cluster import Cluster

# Parameters
from .parameters import PrintParams , NewParams

# Useful function for stellar evolution stuff
from .stellarevo import StarEvo , Value , LoadTrack

# Stellar evolution basic quantities
from .stellarevo import ( Rstar , Lbol , Teff , Itotal , Icore , Ienv , Mcore , Menv , Rcore , tauConv ,
                         dItotaldt , dIcoredt , dIenvdt , dIenvdt , dMcoredt , dRcoredt )

# Functions from physical model
from .physicalmodel import ( dOmegadt , RotationQuantities , ExtendedQuantities , Lxuv , Lx , Leuv , Lly , 
                            OmegaSat , ProtSat , MdotFactor , OmegaBreak , XrayScatter , XUVScatter , aOrbHZ )

# Rotational evolution stuff
from .rotevo import EvolveRotation , EvolveRotationStep

# Spectral synthesis 
from .synthesis import *

# Data download
from .data import DownloadEvolutionTracks

# Baraffe tracks
from .baraffe import *

# Spectrum stuff 
from .spectrum import *

# Some other stuff
from .miscellaneous import Load , ModelCluster , ActivityLifetime , IntegrateEmission



import os
import subprocess
from osfclient.api import OSF

#project ID of the stellar evolution tracks folder in the OSF
project_id = '9u3fb'

def download_folder(storage, folder_name, local_path):
    ''''
    Download a specific folder in the OSF repository

    Inputs :
        - storage     : OSF storage name
        - folder_name : folder name to be downloaded
        - local_path  : local repository where data are saved
    '''
    for file in storage.files:
        if file.path.startswith(folder_name):
            local_file_path = local_path + file.path
            #Create local directory if needed
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)
            #Download the file
            with open(local_file_path, 'wb') as local_file:
                file.write_to(local_file)
    return

def DownloadEvolutionTracks(fname=""):
    '''
    Download evolution track data

    Inputs :
        - fname (optional) :    folder name, "/Spada" or "/Baraffe"
                                if not provided download both
    '''

    #Check if data environment variable is set up
    fwl_data_dir = os.getenv('FWL_DATA')
    if os.environ.get("FWL_DATA") == None:
        raise Exception("The FWL_DATA environment variable where input data will be downloaded needs to be set up!")

    #Create stellar evolution tracks data repository if not existing
    data_dir = fwl_data_dir + "/stellar_evolution_tracks"
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    #Link with OSF project repository
    osf = OSF()
    project = osf.project(project_id)
    storage = project.storage('osfstorage')

    #If no folder name specified download both Spada and Baraffe
    #If local directory exists, assumes the data are already there
    unzip_spada = False
    if not fname:
        if not os.path.exists(data_dir+"/Spada"):
            download_folder(storage,"/Spada",data_dir)
            unzip_spada = True
        if not os.path.exists(data_dir+"/Baraffe"):
            download_folder(storage,"/Baraffe",data_dir)
    elif fname == "/Spada":
        if not os.path.exists(data_dir+"/Spada"):
            download_folder(storage,"/Spada",data_dir)
            unzip_spada = True
    elif fname == "/Baraffe":
        if not os.path.exists(data_dir+"/Baraffe"):
            download_folder(storage,"/Baraffe",data_dir)
    else:
        print("Unrecognised folder name in DownloadEvolutionTrackst st")

    if unzip_spada:
        #Unzip Spada evolution tracks
        wrk_dir = os.getcwd()
        os.chdir(data_dir + '/Spada')
        subprocess.call( ['tar','xvfz', 'fs255_grid.tar.gz'] )
        subprocess.call( ['rm','-f', 'fs255_grid.tar.gz'] )
        os.chdir(wrk_dir)

    return



"""Module for deriving stellar spectra from band-integrated fluxes"""

# Import system libraries 
import numpy as np
import os 

# Import MORS files 
import mors.constants as const
import mors.miscellaneous as misc

# Spectral bands for stellar fluxes, in nm
bands_limits = {
    "xr" : [0.517 , 12.5],      # X-ray,      defined by Mors
    "e1" : [10.0  , 32.0],      # EUV1,       defined by Mors
    "e2" : [32.0  , 92.0],      # EUV2,       defined by Mors
    "uv" : [92.0  , 400.0],     # UV,         defined by Harrison
    "pl" : [400.0 , 1.0e9],     # planckian,  defined by Harrison
    'bo' : [1.e-3 , 1.0e9]      # bolometric, all wavelengths
}

bands_ascending = ["xr","e1","e2","uv","pl"]

def WhichBand(wl:float):
    """Determine which band(s) this wavelength is inside of
    
    Parameters
    ----------
        wl : float
            Wavelength to query [nm]
    
    Returns 
    ----------
        bands : list | None
            List of band names (strings) which this band is inside of. Bands 
            can overlap, so this list may have a length greater than one.     
            If `wl` is outside all bandpasses, then this value is None.   
    
    """

    # Find band, returning when found 
    bands = []
    for b in bands_ascending:
        if bands_limits[b][0] <= wl < bands_limits[b][1]:
            bands.append(b)
    
    # Not found...
    if len(bands) == 0:
        return None 
    
    # Else...
    return bands

class Spectrum():

    def __init__(self):

        # Flags
        self.loaded = False 

        # Arrays (scaled to 1 AU)
        self.nbins = 0      # Number of bins 
        self.wl = []        # Wavelength bins [nm]
        self.fl = []        # Flux bins [erg s-1 cm-2 nm-1]
        self.binwidth = []  # Width of wavelength bins [nm]

        # Extensions
        self.ext_long   = -1 # Index where Planck function extension starts
        self.ext_short  = -1 # Index where shortwave extension starts

        # Integrated fluxes for each band [erg s-1 cm-2]
        self.fl_integ = {}
        for b in bands_limits.keys():
            self.fl_integ[b] = 0.0


    def CalcBandFluxes(self):
        """Calculate integrated fluxes for each band.

        Integrated fluxes will have units of [erg s-1 cm-2] scaled to 1 AU.
        """

        # For each spectral band 
        idxs = []
        i_lo = 0
        for b in bands_ascending:

            # Get band indicies 
            for i in range(i_lo, self.nbins, 1):
                wb = WhichBand(self.wl[i])

                # Out of range
                if wb == None:
                    continue 

                if b in wb:
                    # In current band 
                    idxs.append(i)
                else:
                    # End of band 
                    i_lo = i
                    break 
            
            # With idxs defined, integrate over band 
            band_wl = self.wl[idxs]
            band_fl = self.fl[idxs]
            self.fl_integ[b] = np.trapz(band_fl,band_wl)

            # Reset idxs 
            # print("%s band between in interval [%.2f, %.2f] nm"%(b,self.wl[idxs[0]],self.wl[idxs[-1]]))
            idxs = []

        # For bolometric "band"
        self.fl_integ["bo"] = np.trapz(self.fl,self.wl)

        return self.fl_integ

    def LoadDirectly(self, spec_wl:np.ndarray, spec_fl:np.ndarray):
        """Store spectral data in object.

        Scaled to 1 AU. 

        Parameters
        ----------
            spec_wl : np.ndarray
                Array of wavelengths [nm] 
            spec_fl : np.ndarray
                Array of fluxes [erg s-1 cm-2 nm-1]
        """


        # Check length 
        if len(spec_wl) != len(spec_fl):
            misc._PrintErrorKill("Stellar spectrum size mismatch (%d and %d)"%(len(spec_wl), len(spec_fl)))
        if len(spec_wl) < 10:
            misc._PrintErrorKill("Stellar spectrum size too small (%d bins)"%len(spec_wl))

        # Check reversal (should be wl ascending)
        if spec_wl[4] < spec_wl[0]:
            spec_wl = spec_wl[::-1]
            spec_fl = spec_fl[::-1]

        # Replace NaN and zero values 
        for i in range(len(spec_fl)):
            if not np.isfinite(spec_fl[i]):
                spec_fl[i] = 0.0
            spec_fl[i] = max(spec_fl[i], 1e-20)

        # Calculate bin width
        binwidth_wl = spec_wl[1:] - spec_wl[0:-1]

        # Store 
        self.nbins = len(spec_wl)
        self.wl = spec_wl
        self.fl = spec_fl 
        self.binwidth = np.array(binwidth_wl, dtype=float)

        self.loaded = True

        return self 


    def LoadTSV(self, fp:str):
        """Load stellar spectrum from TSV file into memory.

        Scaled to 1 AU. File should be whitespace delimited with 
        units of [nm] and [erg s-1 cm-2 nm-1].

        Parameters
        ----------
            fp : str
                Path to file 
        """

        # Check path
        fp = os.path.abspath(fp)
        if not os.path.isfile(fp):
            misc._PrintErrorKill("Cannot find TSV file at '%s'"%fp)

        # Load file
        spec_data = np.loadtxt(fp).T
        spec_wl = spec_data[0]
        spec_fl = spec_data[1]

        # Process data 
        spec_wl = np.array(spec_wl, dtype=float)
        spec_fl = np.array(spec_fl, dtype=float)

        # Store 
        self.LoadDirectly(spec_wl, spec_fl)

        self.loaded = True 

        return self 


    def ExtendShortwave(self, wl_min:float):
        """Extend spectrum to shorter wavelengths using constant value.

        Parameters
        ----------
            wl_min : float 
                New minimum wavelength [nm]
        """

        # Already extended
        if wl_min > self.wl[0]:
            return 
        
        # Calc wavelength extension
        wl_ext = np.logspace(np.log10(wl_min), np.log10(self.wl[0]), 200)[:-1]

        # Evalulate planck function
        fl_ext = np.ones(np.shape(wl_ext)) * self.fl[0]

        # Update Spectrum object data 
        self.ext_short = len(wl_ext)
        spec_wl = np.concatenate((wl_ext,self.wl))
        spec_fl = np.concatenate((fl_ext,self.fl))

        # Store 
        self.LoadDirectly(spec_wl, spec_fl)


    def ExtendPlanck(self, Teff:float, R_star:float, wl_max:float):
        """Extend spectrum to longer wavelengths using planck function.

        Parameters
        ----------
            Teff : float 
                Effective temperature of star 
            R_star : float 
                Radius of star [m]
            wl_max : float 
                New maximum wavelength [nm]
        """


        # Already extended
        if wl_max < self.wl[-1]:
            return 
        
        # Calc wavelength extension
        wl_ext = np.logspace(np.log10(self.wl[-1]), np.log10(wl_max), 300)[1:]

        # Evalulate planck function
        fl_ext = PlanckFunction_surf(wl_ext, Teff)

        # Scale to 1 AU
        fl_ext = ScaleTo1AU(fl_ext, R_star)

        # Update Spectrum object data 
        self.ext_long = len(self.wl)
        spec_wl = np.concatenate((self.wl, wl_ext))
        spec_fl = np.concatenate((self.fl, fl_ext))

        # Store 
        self.LoadDirectly(spec_wl, spec_fl)

    
    def WriteTSV(self, fp:str):
        """Write spectrum to file(s) on disk.

        Parameters
        ----------
            fp : str
                Path to file 
        """

        fp = os.path.abspath(fp)
        X = np.array([self.wl,self.fl]).T
        header = "WL(nm)\t Flux(ergs/cm**2/s/nm)    Stellar flux (1 AU)"

        np.savetxt(fp, X, header=header,fmt='%1.4e',delimiter='\t')

        return fp


def PlanckFunction_surf(wl:np.ndarray, Teff:float):
    """Returns the planck fluxes evaluated at the wavelength array
    
    Parameters
    ----------
        wl : np.ndarray
            Wavelength array [nm]
        Teff : float
            Effective temperature of the object

    Returns
    ----------
        yp : float
            Flux at stellar surface [erg s-1 cm-2 nm-1]
    """

    yp = np.zeros(np.shape(wl))
    for i,x in enumerate(wl):
        lam = x * 1.0e-9  # nm -> m

        # Calculate planck function value [W m-2 sr-1 m-1]
        # http://spiff.rit.edu/classes/phys317/lectures/planck.html
        yp[i] = 2.0 * const.h_SI * const.c_SI**2.0 / lam**5.0   *   1.0 / ( np.exp(const.h_SI * const.c_SI / (lam * const.k_SI * Teff)) - 1.0)

        # Integrate solid angle (hemisphere), convert units
        yp[i] = yp[i] * np.pi * 1.0e-9 # [W m-2 nm-1]
        yp[i] = yp[i] * 1000.0 # [erg s-1 cm-2 nm-1]

    return yp


def ScaleToSurf(fl:np.ndarray, R_star:float):
    """Scale spectrum from 1 AU to stellar surface
    
    Parameters
    ----------
        fl : np.ndarray
            Spectrum at 1 AU
        R_star : float
            Star radius in [m]

    Returns
    ----------
        fl_surf : np.ndarray
            Flux at stellar surface (same units as `fl`)
    """

    return fl * (const.AU_SI/R_star)**2


def ScaleTo1AU(fl:np.ndarray, R_star:float):
    """Scale spectrum from stellar surface to 1AU
    
    Parameters
    ----------
        fl : np.ndarray
            Spectrum at surface
        R_star : float
            Star radius in [m]

    Returns
    ----------
        fl_1AU : np.ndarray
            Flux at 1 AU (same units as `fl`)
    """

    return fl * (R_star/const.AU_SI)**2




"""Module for loading and interpolating Baraffe tracks data.
Original tracks data can be found on the website 
http://perso.ens-lyon.fr/isabelle.baraffe/BHAC15dir/BHAC15_tracks+structure"""
import numpy as np
import os
from scipy.interpolate import PchipInterpolator

import mors.constants as const

#Short cut to Baraffe tracks mass and temporal range
MassGrid = [0.010, 0.015, 0.020, 0.030, 0.040, 0.050,
            0.060, 0.070, 0.072, 0.075, 0.080, 0.090,
            0.100, 0.110, 0.130, 0.150, 0.170, 0.200,
            0.300, 0.400, 0.500, 0.600, 0.700, 0.800,
            0.900, 1.000, 1.100, 1.200, 1.300, 1.400 ]
tminGrid = [ 5.695210, 5.693523, 5.689884, 5.695099, 5.694350, 5.694123,
             5.694808, 5.692913, 5.694756, 5.694352, 5.694252, 5.689450,
             5.703957, 5.709729, 5.702708, 5.701659, 5.693846, 5.693823,
             5.690044, 5.689995, 5.694573, 5.690793, 5.691618, 5.693354,
             5.692710, 5.693063, 5.694315, 5.693263, 5.694626, 5.692977 ]
tmaxGrid = [ 7.612341, 8.050550, 8.089757, 8.514580, 8.851534, 9.178493,
             9.499851,10.000078, 9.999692,10.000130,10.000343, 9.999745,
             9.999914, 9.999209, 9.999428,10.000104, 9.999246, 9.999735,
            10.000004, 9.999193, 9.999099, 9.999787, 9.999811, 9.998991,
            10.000065, 9.920501, 9.789029, 9.651666, 9.538982, 9.425490 ]

fwl_data_dir = os.getenv('FWL_DATA')

class BaraffeTrack:
    """Class to hold interpolated baraffe tracks data for a given star mass

    Attributes
    ----------
    mstar : float
        Star mass
    track : dict
        Dictionary containing track data
    """

    def __init__(self, Mstar):

        self.mstar = Mstar

        #If Mstar is out of the mass range, code breaks
        if(Mstar < MassGrid[0]):
            raise Exception("Stellar mass is too low for the Baraffe tracks")
        elif(Mstar > MassGrid[29]):
            raise Exception("Stellar mass is too high for the Baraffe tracks")

        #If Mstar matches with values in the mass grid
        elif(Mstar in MassGrid):
            track = BaraffeLoadTrack(Mstar)

        #If Mstar is between two points in the mass grid, need mass interpolation
        else:
            for i in range(29):
                if not Mstar > MassGrid[i+1]:
                    #search for common temporal range
                    tmin = max(tminGrid[i],tminGrid[i+1])
                    tmax = min(tmaxGrid[i],tmaxGrid[i+1])
                    
                    #load neighbouring tracks with time interpolation on the same time grid
                    track  = BaraffeLoadTrack(MassGrid[i  ], tmin=tmin, tmax=tmax)
                    trackp = BaraffeLoadTrack(MassGrid[i+1], tmin=tmin, tmax=tmax)
                   
                    #perform linear mass interpolation for each array
                    mass_ratio=(Mstar-MassGrid[i])/(MassGrid[i+1]-MassGrid[i])
                    track['Teff' ]=(trackp['Teff' ]-track['Teff' ])*mass_ratio + track['Teff' ]
                    track['Lstar']=(trackp['Lstar']-track['Lstar'])*mass_ratio + track['Lstar']
                    track['Rstar']=(trackp['Rstar']-track['Rstar'])*mass_ratio + track['Rstar']

                    break

        self.track = track
        self.tmin = track['t'][0]
        self.tmax = track['t'][-1]

        return

    def BaraffeLuminosity(self, tstar):
        """Calculates the star luminosity at a given time.

        Parameters
        ----------
            tstar : float
                Star's age

        Returns
        ----------
            Lstar : float
                Luminosity Flux in units of solar luminosity
        """

        # Get time and check that it is in range
        if (tstar < self.tmin):
            print("Star age too low! Clipping to %.1g Myr" % int(self.tmin*1.e-6))
            tstar = self.tmin
        if (tstar > self.tmax):
            print("Star age too high! Clipping to %.1g Myr" % int(self.tmax*1.e-6))
            tstar = self.tmax

        # Find closest row in track
        iclose = (np.abs(self.track['t'] - tstar)).argmin()

        # Get data from track
        Lstar = self.track['Lstar'][iclose]

        return Lstar

    def BaraffeSolarConstant(self, tstar, mean_distance):
        """Calculates the bolometric flux of the star at a given time.
        Flux is scaled to the star-planet distance.

        Parameters
        ----------
            tstar : float
                Star's age
            mean_distance : float
                Star-planet distance

        Returns
        ----------
            inst : float
                Flux at planet's orbital separation (solar constant) in W/m^2
        """

        Lstar = self.BaraffeLuminosity(tstar)
        Lstar *= const.LbolSun_SI
        mean_distance *= const.AU_SI

        inst = Lstar / ( 4. * np.pi * mean_distance * mean_distance )

        return inst

    def BaraffeStellarRadius(self, tstar):
        """Calculates the star's radius at a time t.

        Parameters
        ----------
            tstar : float
                Star's age

        Returns
        ----------
            Rstar : float
                Radius of star in units of solar radius
        """

        # Get time and check that it is in range
        if (tstar < self.tmin):
            print("Star age too low! Clipping to %.1g Myr" % int(self.tmin*1.e-6))
            tstar = self.tmin
        if (tstar > self.tmax):
            print("Star age too high! Clipping to %.1g Myr" % int(self.tmax*1.e-6))
            tstar = self.tmax

        # Find closest row in track
        iclose = (np.abs(self.track['t'] - tstar)).argmin()

        # Get data from track
        Rstar = self.track['Rstar'][iclose]

        return Rstar

def BaraffeLoadTrack(Mstar, pre_interp = True, tmin = None, tmax = None):
    """Load a baraffe track into memory and optionally interpolate it into a fine time-grid

    Parameters
    ----------
        Mstar : float
            Star mass (in unit of solar mass)
            It assumes the value has been checked and matches with the mass grid.
        pre_interp : bool
            Pre-interpolate the tracks onto a higher resolution time-grid
        tmin : float
            Minimum value of the temporal grid
        tmax : float
            Maximum value of the temporal grid

    Returns
    ----------
        track : dict
            Dictionary containing track data
    """

    # Load data
    formatted_mass = f"{Mstar:.3f}"
    file = (fwl_data_dir +
           "/stellar_evolution_tracks/Baraffe/BHAC15-M" +
           str(formatted_mass).replace('.', 'p') +
           ".txt")
    if not os.path.isfile(file):
        raise Exception(
            "Cannot find Baraffe track file %s. "
            "Did you set the FWL_DATA environment variable? "
            "Did you run the DownloadEvolutionTracks function to get access to the Baraffe track data?"
            % file
        )

    data = np.loadtxt(file,skiprows=1).T

    # Parse data
    t =     data[1]
    Teff =  data[2]
    Lstar = data[3]
    Rstar = data[5]

    # Pre-interpolate in time if required
    if pre_interp:
        # Params for interpolation
        grid_count = 5e4
        if tmin==None: tmin = t[0]
        if tmax==None: tmax = t[-1]

        # Do interpolation
        #log.info("Interpolating stellar track onto a grid of size %d" % grid_count)
        interp_Teff =   PchipInterpolator(t,Teff)
        interp_Lstar =  PchipInterpolator(t,Lstar)
        interp_Rstar =  PchipInterpolator(t,Rstar)

        new_t = np.logspace(tmin, tmax, int(grid_count))
        new_t = np.log10(new_t)
        new_Teff =  interp_Teff(new_t)
        new_Lstar = interp_Lstar(new_t)
        new_Rstar = interp_Rstar(new_t)

        track = {
            't':        10.0**new_t,      # yr
            'Teff':     new_Teff,         # K
            'Lstar':    10.0**new_Lstar,  # L_sun
            'Rstar':    new_Rstar         # R_sun
        }
    else:
        track = {
            't':        10.0**t,      # yr
            'Teff':     Teff,         # K
            'Lstar':    10.0**Lstar,  # L_sun
            'Rstar':    Rstar         # R_sun
        }
        
    return track




"""Module for holding default parameters and related functions."""

# Imports for standard stuff needed here
import copy
import numpy as np

# Make empty dictionary to hold default values for all parameters in the code. Some of these
# are parameters that the user might want to change in order to determine the behavior of the code.
paramsDefault = {}

def SetDefaultParameters(paramsDefault):
    
    """Sets up the default parameters."""
    
    # Time integration solver parameters
    paramsDefault['TimeIntegrationMethod'] = 'Rosenbrock'     # options are 'ForwardEuler', 'RungeKutta4', 'RungeKuttaFehlberg', 'Rosenbrock'
    paramsDefault['nStepMax'] = 10**6                         # maximum number of timesteps to allow before error
    paramsDefault['DeltaDesired'] = 1.0e-5                    # desired Delta value to use in numerical solvers such as Runge-Kutta-Fehlberg and Rosenbrock
    paramsDefault['AgeMinDefault'] = 1.0                      # Myr - default age to start evolutionary tracks
    paramsDefault['AgeMaxDefault'] = 5000.0                   # Myr - default age to end evolutionary tracks
    paramsDefault['deltaJac'] = 1.0e-5                        # pertubation to use when calculating Jacobian for Rosenbrock solver
    paramsDefault['CoefficientsRB'] = _CoefficientsRB()       # all coefficents needed for the Rosenbrock solver
    paramsDefault['dAgeMin'] = 1.0e-5                         # Myr - minimum timestep to allow
    paramsDefault['dAgeMax'] = 5000.0                         # Myr - maximum timestep to allow
    
    # Physical processes to include
    paramsDefault['CoreEnvelopeDecoupling'] = True            # should core envelope decoupling be included
    paramsDefault['MomentInertiaChangeTorque'] = True         # should the code include effects of moment of inertia changes on rotation (do not turn this off)
    paramsDefault['CoreGrowthTorque'] = True                  # should the code include effects of core-growth on core and envelope rotation (do not turn this off)
    paramsDefault['WindTorque'] = True                        # should the code include wind spin-down torque
    paramsDefault['CoreEnvelopeTorque'] = True                # should the code include core-envelope coupling torque
    paramsDefault['DiskLocking'] = True                       # should disk-locking be used
    
    # Core envelope decoupling parameters
    paramsDefault['IcoreThresholdCE'] = 0.01                  # value of Icore/Itotal below which core-envelope decoupling is not used
    paramsDefault['MstarThresholdCE'] = 0.35                  # Msun - value of Mstar below which core-envelope decoupling is not used
    
    # Disk locking parameters
    paramsDefault['aDiskLock'] = 13.5                         # a in tDiskLock = a * Omega^b
    paramsDefault['bDiskLock'] = -0.5                         # b in tDiskLock = a * Omega^b
    paramsDefault['ageDLmax'] = 15.0                          # Myr - maximum disk locking age
    
    # Wind parameters
    paramsDefault['Kwind'] = 11.0                             # additional constant in torque formula
    paramsDefault['BdipSun'] = 1.35                           # G - estimate for rotation models from Johnstone et al. (2015)
    paramsDefault['aBdip'] = -1.32                            # from Vidotto et al. (2014)
    paramsDefault['RoSatBdip'] = 0.0605                       # saturation Rossby number for dipole field strength from Johnstone et al. (2020) using Spada et al. (2013) models
    paramsDefault['MdotSun'] = 1.4e-14                        # Msun yr^-1 - solar mass loss rate
    paramsDefault['aMdot'] = -1.7591                          # power law index on Ro in mass loss rate equation from Johnstone et al. (2020)
    paramsDefault['bMdot'] = 0.6494                           # power law index on Mstar in mass loss rate equation from Johnstone et al. (2020)
    paramsDefault['RoSatMdot'] = 0.0605                       # saturation Rossby number for mass loss rate from Johnstone et al. (2020) using Spada et al. (2013) models
    paramsDefault['BreakupMdotIncrease'] = True               # should we increase Mdot as approaching breakup
    paramsDefault['fracBreakThreshold'] = 0.1                 # fraction of breakup to start increasing Mdot
    
    # Core-envelope coupling stuff
    paramsDefault['aCoreEnvelope'] = 25.6015                  # a in tCoreEnvelope = a * (|OmegaCore-OmegaEnv|)^b * Mstar^d
    paramsDefault['bCoreEnvelope'] = -3.4817E-002             # b in tCoreEnvelope = a * (|OmegaCore-OmegaEnv|)^b * Mstar^d
    paramsDefault['cCoreEnvelope'] = -0.4476                  # c in tCoreEnvelope = a * (|OmegaCore-OmegaEnv|)^b * Mstar^d
    paramsDefault['timeCEmin'] = 0.0                          # minimum value of core-envelope coupling timescale
    
    # Output parameters
    paramsDefault['ExtendedTracks'] = False                   # should the code return evolutionary tracks for all quantities 
    
    # XUV emission parameters
    paramsDefault['RoSatXray'] = 0.0605                       # saturation Rossby number for X-ray emission from Johnstone et al. (2020) using models of Spada et al. (2013)
    paramsDefault['RxSatXray'] = 5.135e-4                     # saturation Rx
    paramsDefault['beta1Xray'] = -0.135                       # beta from Rx = C * Ro^beta for the saturated regime
    paramsDefault['beta2Xray'] = -1.889                       # beta from Rx = C * Ro^beta for the unsaturated regime
    paramsDefault['C1Xray'] = paramsDefault['RxSatXray'] / paramsDefault['RoSatXray']**paramsDefault['beta1Xray'] # C from Rx = C * Ro^beta for the saturated regime
    paramsDefault['C2Xray'] = paramsDefault['RxSatXray'] / paramsDefault['RoSatXray']**paramsDefault['beta2Xray'] # C from Rx = C * Ro^beta for the unsaturated regime
    paramsDefault['sigmaXray'] = 0.359                        # dex - standard deviation of normal PDF for X-ray scatter
    
    # Rotation fitting parameters
    paramsDefault['AgeMinFit'] = 1.0                          # Myr - starting age for fitting rotation track
    paramsDefault['Omega0FitMin'] = 0.1                       # OmegaSun - minimum starting rotation rate to consider when fitting initial rotation
    paramsDefault['Omega0FitMax'] = 50.0                      # OmegaSun - maximum starting rotation rate to consider when fitting initial rotation
    paramsDefault['nStepMaxFit'] = 1000                       # maximum number of steps to take when fitting rotation rate
    paramsDefault['toleranceFit'] = 1.0e-5                    # tolerance for fitting initial rotation rate
    
    # Other
    paramsDefault['dMstarPer'] = 0.1                          # Msun - half width of bins for working out percentiles of distribution
    paramsDefault['AgeHZ'] = 5000.0                           # Myr - age to use for calculating habitable zone boundaries
    
    return paramsDefault

def _CoefficientsRB():
    
    """Returns dictionary with basic coefficients for Rosenbrock solver."""

    # Using here values from RODAS3 of Sandu et al. (1997)
    
    # Set s value (i.e. number of steps)
    s = 4
    
    # Set alpha_ij
    alpha = np.zeros((s,s))
    alpha[2,0] = 1.0
    alpha[3,0] = 3.0/4.0
    alpha[3,1] = -1.0/4.0
    alpha[3,2] = 1.0/2.0
    
    # Set gamma_ij
    gamma = np.zeros((s,s))
    gamma[0,0] = 1.0/2.0
    gamma[1,0] = 1.0
    gamma[2,0] = -1.0/4.0
    gamma[3,0] = 1.0/12.0
    gamma[1,1] = 1.0/2.0
    gamma[2,1] = -1.0/4.0
    gamma[3,1] = 1.0/12.0
    gamma[2,2] = 1.0/2.0
    gamma[3,2] = -2.0/3.0
    gamma[3,3] = 1.0/2.0

    # Set b_i
    b = np.zeros(s)
    b[0] = 5.0/6.0
    b[1] = -1.0/6.0
    b[2] = -1.0/6.0
    b[3] = 1.0/2.0
    
    # Set b2_i
    b2 = np.zeros(s)
    b2[0] = 3.0/4.0
    b2[1] = -1.0/4.0
    b2[2] = 1.0/2.0
    b2[3] = 0.0
    
    # Set the order of the scheme
    order = 3
    
    # Put these into a dictionary for returning
    CoefficientsRB = {}
    CoefficientsRB['s'] = s
    CoefficientsRB['alpha'] = alpha
    CoefficientsRB['gamma'] = gamma
    CoefficientsRB['b'] = b
    CoefficientsRB['b2'] = b2
    CoefficientsRB['order'] = order
    
    return CoefficientsRB

def PrintParams(params=paramsDefault):
    
    """Takes parameter dictionary, prints values to screen."""
    
    # Loop over parameters and print each
    for param in params:
        print(param , " = " , params[param] )
    
    return

def NewParams(**kwargs):
    
    """Takes list of parameters, returns new parameter dictionary using defaults for non-specified parameters."""
    
    # Make new parameter dictionary based on paramsDefault
    params = copy.deepcopy(paramsDefault)
    
    # Change parameters based on user specifications
    for param in kwargs:
        params[param] = kwargs[param]
    
    return params

# Setup the default parameter dictionary
paramsDefault = SetDefaultParameters(paramsDefault)




"""Module for holding miscellaneous functions that can be used by many other modules."""

# Imports for standard stuff needed here
import inspect
import sys
import numpy as np
import pickle
import copy
import pathlib

# Imports for mors modules
import mors.constants as const

def _PrintErrorKill(errorString):
    """Takes string with error message, prints message to screen and end code execution."""
    
    # Get info about the calling function
    frame = inspect.stack()[1]
    filename = frame.filename
    function = frame.function
    
    # From filename, just get relative to kompot/ or kvass/ directory
    temp = filename.split('/')
    filename = temp[-1]
    
    # Print error to screen
    print( "Error in "+function+"() in file "+filename+": "+errorString )
    
    # Stop the code
    sys.tracebacklimit = 1
    raise ValueError()
    
    return

def _PrintError(errorString):
    """Takes string with error message, prints message to screen."""

    # Get info about the calling function
    frame = inspect.stack()[1]
    filename = frame.filename
    function = frame.function
    
    # From filename, just get relative to kompot/ directory
    temp = filename.split('/')
    filename = temp[-1]
    
    # Print error to screen
    print( "Error in "+function+"() in file "+filename+": "+errorString )
    
    return

def _GetPackageDirectory():
    """Gets path to main directory where the code is installed."""
    
    return str(pathlib.Path(__file__).parent.absolute())+"/data/"

def _convertFloatArray(Xin):
    """Takes a value and returns either as float or numpy.ndarray of floats."""
    
    # If input is float or int, return float version
    if isinstance(Xin,(float,int)):
        return float(Xin)
    
    # If input is a list, numpy array version (will only work if elements can be converted into floats)
    if isinstance(Xin,list):
        X = np.zeros(len(Xin))
        for i in range(0,len(Xin)):
            X[i] = float(Xin[i])
        return X
    
    # If input is a numpy object then do stuff
    if ( type(Xin).__module__ == 'numpy' ):
        
        # First see if it can be converted into a float (i.e. a 1-element array)
        isNumber = True
        try:
            X = float(Xin)
        except:
            isNumber = False
        
        # If it is just a single value, return float version
        if ( isNumber ):
            return float(X)
        
        # Otherwise, turn it into an array
        X = np.zeros(len(Xin))
        for i in range(0,len(Xin)):
            X[i] = float(Xin[i])
        return X
    
    return

def Load(filename):
    """Takes filename of saved star or cluster and loads object."""
    
    # Load the object
    with open(filename,'rb') as f:
        obj = pickle.load(f)
    
    # Setup the quantity functions (which for some reason do not work on objects loaded from pickle)
    obj._setupQuantityFunctions()
    
    return obj

def ModelCluster():
    """Reads the model cluster used in Johnstone et al. (2020)."""
    
    # Get directory where package is installed
    packageDir = _GetPackageDirectory()
    
    # Set filename including path
    filename = packageDir + "ModelDistribution.dat"
    
    # Read lines of file
    with open(filename) as f:
        content = f.readlines()
    
    # Set number of header lines
    nHeader = 1
    
    # Get number of steps in file
    nStars = len(content) - nHeader
    
    # Make arrays
    Mstar = np.zeros(nStars)
    Omega = np.zeros(nStars)
    
    # Loop over lines and fill arrays
    iStar = 0
    for line in content[nHeader:len(content)]:
        data = line.split()
        Mstar[iStar] = data[0]
        Omega[iStar] = data[1]
        iStar += 1
    
    return Mstar , Omega

def ActivityLifetime(Age=None,Track=None,Threshold=None,AgeMax=None):
    """
    Takes evolutionary track for parameter, calculates when value drops below threshold.

    This function can be used to determine when a star's emission crosses a given threshold value for a few
    activity quantities. These are Lx, Fx, Rx, and FxHZ for X-rays, and similar values for EUV1, EUV2, EUV, 
    XUV, and Ly-alpha. If the star crosses the threshold (from above it to below it) multiple times, this 
    will find the final time it will cross the threshold. If the user wants to set a maximum age so that the 
    code only looks for crossings of the threshold below this age then this can be done using the AgeMax
    keyword argument. If the track is always below this threshold value then the function returns 0.0.
        
    Parameters
    ----------
    Age : numpy.ndarray
        Age array for evolutionary track.
    Track : numpy.ndarray
        Value array for evolutionary track.
    Threshold : float
        Value for threshold in units of quantity.
    AgeMax : float , optional
        End age of track to consider in Myr.
    
    Returns
    ----------
    AgeActive : float
        Activity lifetime in Myr.
    
    """
  
    # Make sure parameters set
    if Age is None:
        misc._PrintErrorKill("required argument Age not set")
    if Track is None:
        misc._PrintErrorKill("required argument Track not set")
    if Threshold is None:
        misc._PrintErrorKill("required argument Threshold not set")
        
    # Make sure Age and Track same length
    if not ( len(Age) == len(Track) ):
        misc._PrintErrorKill("Age and Track are different lengths")
    
    # Change Age and Track to only be ages below AgeMax, if AgeMax is set
    if not AgeMax is None:
        includeAges = np.where( Age <= AgeMax )
        Age = Age[includeAges]
        Track = Track[includeAges]
        
    # See if final value below threshold, otherwise return final age
    if ( Track[-1] > Threshold ):
        return Age[-1]
        
    # Start at end of array and loop backwards
    for iAge in range(len(Age)-1,0,-1):
        
        # Make sure next one back is not equal to threshold
        if ( Track[iAge-1] == Threshold ):
            return Age[iAge-1]
        
        # Check if this age bin is when it drops past the threshold 
        if ( ( Track[iAge] < Threshold ) and ( Track[iAge-1] > Threshold ) ):
            
            # Do interpoaltion to get when it crosses 
            # Assume here Age = m*Track + c
            mInterp = ( Age[iAge] - Age[iAge-1] ) / ( Track[iAge] - Track[iAge-1] )
            cInterp = Age[iAge] - mInterp * Track[iAge]
            return mInterp*Threshold + cInterp
      
    return 0.0

def IntegrateEmission(AgeMin=None,AgeMax=None,Age=None,Luminosity=None,aOrb=None):
    """
    Takes evolutionary track for parameter, calculates when value drops below threshold.
    
    This code can be used to integrate a star's luminosity between two ages. This can be applied to Lx, Leuv, or any
    other luminosity and the result is a total energy emitted in this time in erg. If the user also specifies an orbital 
    distance using the aOrb keyword argument, the code integrates the flux at this obital distance returns a fluence 
    in erg cm^-2.
        
    Parameters
    ----------
    AgeMin : float
        Start of time period to integrate in Myr.
    AgeMax : float
        End of time period to integrate in Myr.
    Age : numpy.ndarray
        Age array for evolutionary track.
    Luminosity : numpy.ndarray
        Value array for evolutionary track.
    aOrb : float , optional
        Orbital distance to get fluence at in AU.
    
    Returns
    ----------
    Energy : float
        Integrated luminosity of flux in erg or erg cm^-2.
    
    """
    
    # Make sure parameters set
    if AgeMin is None:
        misc._PrintErrorKill("required argument AgeMin not set")
    if AgeMax is None:
        misc._PrintErrorKill("required argument AgeMax not set")
    if Age is None:
        misc._PrintErrorKill("required argument Age not set")
    if Luminosity is None:
        misc._PrintErrorKill("required argument Luminosity not set")
        
    # Make sure Age and Luminosity same length
    if not ( len(Age) == len(Luminosity) ):
        misc._PrintErrorKill("Age and Luminosity are different lengths")
    
    # Make sure AgeMin and AgeMax is in track
    if not ( ( AgeMin >= Age[0] ) and ( AgeMin <= Age[-1] ) ):
        misc._PrintErrorKill("AgeMin not in range of evolutionary track")
    if not ( ( AgeMax >= Age[0] ) and ( AgeMax <= Age[-1] ) ):
        misc._PrintErrorKill("AgeMax not in range of evolutionary track")
    
    # Get track to actually integrate
    Track = copy.deepcopy(Luminosity)
    if not aOrb is None:
        Track *= 1.0 / ( 4.0 * const.Pi * (aOrb*const.AU)**2.0 )
    
    # Get index of first age bin with age above AgeMin
    indexMin = SE._getIndexGT(Age,AgeMin)
    
    # Get index of final age bin with age below AgeMax
    indexMax = SE._getIndexLT(Age,AgeMax)
    
    # Initially take no energy then add up energy from bins
    Energy = 0.0
    
    # Add energy from first bin if needed (simple assumption for luminosity constant ove age bin)
    if ( AgeMin < Age[indexMin] ):
        Energy += ( Age[indexMin] - AgeMin ) * Track[indexMin]
    
    # Intergate over all full bins (only if there is something to integrate)
    if not ( indexMin == indexMax ):
        for iAge in range(indexMin,indexMax):
            
            # Do integration using trapezoidal rule (not the chained version)
            Energy += 0.5 * ( Age[iAge+1] - Age[iAge] ) * ( Track[iAge] + Track[iAge+1] )
        
    # Add energy from final bin if needed (simple assumption for luminosity constant ove age bin)
    if ( AgeMax > Age[indexMax] ):
        Energy += ( AgeMax - Age[indexMax] ) * Track[indexMax]
    
    # So far, units of Age were in Myr when integrating, so include also Myr to s conversion
    Energy *= const.Myr
    
    return Energy

def _getIndexLTordered(Xarray,X):
    """Takes min to max ordered array of values and a value, returns index of closest element in array smaller than value."""
    
    # It is assumed that X is between the min and max of Xarray and it is assumed
    # that Xarray is in ascending order
        
    # Initial guesses for i
    i1 = 0
    i2 = len(Xarray)-1
    
    # Check if right value
    if ( Xarray[i1] == X ):
        return i1
    if ( Xarray[i2] == X ):
        return i2
        
    # Start iterating
    for iIter in range(0,len(Xarray)):
        
        # Get iMid
        iMid = int(0.5*(i1+i2))
        
        # Check if iMid is right value
        if ( ( Xarray[iMid] <= X ) and ( Xarray[iMid+1] > X ) ):
            return iMid
        
        # Work out if answer is between i1 and iMid or iMid and i2
        if ( Xarray[iMid] > X ):
            i2 = iMid
        else:
            i1 = iMid
    
    # It should not get here
    misc._PrintErrorKill("did not find index")
    
    return iMid

def _getIndexLT(Xarray,X):
    """Takes array of values and a value, returns index of closest element in array smaller than value."""
    
    # It is assumed that X is greater than the min of Xarray, but check this
    if ( X < np.min(Xarray) ):
        misc._PrintErrorKill("X is less than minimum of Xarray")
    
    # Get smaller value by taking difference of X and all Xarray elements
    # then removing all values smaller than 0.0 (i.e. those larger than X)
    deltaX = X - Xarray
    deltaX[np.where(deltaX<0.0)] = np.max(deltaX)*1.1+0.1
    
    # Get index
    index = np.argmin(deltaX)
    
    return index

def _getIndexGT(Xarray,X):
    """Takes array of values and a value, returns index of closest element in array larger than value."""
    
    # It is assumed that X is less than the max of Xarray, but check this
    if ( X > np.max(Xarray) ):
        misc._PrintErrorKill("X is more than maximum of Xarray")
    
    # Get smaller value by taking difference of X and all Xarray elements
    # then removing all values smaller than 0.0 (i.e. those larger than X)
    deltaX = Xarray - X
    deltaX[np.where(deltaX<0.0)] = np.max(deltaX)*1.1+0.1
    
    # Get index
    index = np.argmin(deltaX)
    
    return index




"""Module for holding some basic constants."""

# Time units
sec = 1.0                   # seconds
minute = 60.0*sec             # 
hour = 60.0*minute        # 
day = 24.0*hour           # 
year = 365.24*day
Myr = 1.0e6*year
Gyr = 1.0e3*Myr

# Distance units in cm
A = 1.0e-8                # Angstrom
nm = 1.0e-7               # nanometer
micron = 1.0e-4           # micrometer
mm = 0.1                  # milimeter
cm = 1.0                  # centimeter
km = 1.0e5                # kilometer
m = 1.0e3                 # meter
Rearth = 6.371e8          # Radius of Earth
Rjup = 6.9911e9           # Radius of Jupiter
Rsun = 6.957e10           # cm
AU = 1.496e13             # cm

# Area units
cm2 = 1.0                 # cm^2

# Mass units in g
g = 1.0                   # gram
kg = 1.0e3                # kilogram
Mearth = 5.972e27         # Mass of Earth
Mjup = 1.89813e30         # Mass of Juputer
Msun = 1.99e33            # Mass of Sun
Mproton = 1.6726219e-24   # Mass of proton
Melec = 9.10938356e-28    # Mass of electron

# Thermal units
K = 1.0                   # Kelvin
erg = 1.0                 # erg
ergs_cm_3 = 1.0           # erg s^-1 cm^-3

# Densities
cm_3 = 1.0                # cm^-3

# Speeds
kms_ = 1.0e-5             # km s^-1 in cm s^-1

# Radiation units
ergs_cm_2 = 1.0           # erg s^-1 cm^-2

# Some other things
Msunyr_ = Msun / year      # g s^-1

# Other solar quantities
LbolSun = 3.828e33        # erg s^-1
LbolSun_SI = 3.828e26     # W
AgeSun = 4567.0           # Myr
OmegaSun = 2.67e-6        # rad s^-1 - our assumed solar rotation
ProtSun = 2.0*3.142 / ( OmegaSun ) / 86400.0    # days - corresponding solar rotation period
tauConvSun = 28.436   # days - the Sun's convective turnover time using stellar evo models of Spada et al. (2013)
RoSun = ProtSun / tauConvSun                    # Sun's current Rossby number using stellar evo models

# Standard constants
GravConstant = 6.674e-8   # cm^3 g^-1 s^-2
kB = 1.38064852e-16       # erg K^-1
Pi = 3.14159265359        # you know what this is

h_SI  = 6.626075540e-34      # J s
c_SI  = 2.99792458e8         # m s^-1
k_SI  = 1.38065812e-23       # J K^-1
AU_SI = 1.495979e+11         # m
